/*
 * CCommandFactory.cpp
 *
 *  Created on: May 24, 2010
 *      Author: pdas
 */

#include "CCommandFactory.h"
// specific command handler header files
#include "CLDACommandHandler.h"
#include "CLDACrossValCommandHandler.h"
#include "CLDACrossValOnlineCommandHandler.h"
#include "CPureSamplingCommandHandler.h"

using namespace std;

/**
 * The command factory creates a map of concrete commands in the constructor
 */
CCommandFactory::CCommandFactory() {
	_command_handler_map.insert( make_pair("lda-dirsym",new CLDACommandHandler()) );
	_command_handler_map.insert( make_pair("lda-dirsym-cross-validate", new CLDACrossValCommandHandler()) );
	_command_handler_map.insert( make_pair("lda-dirsym-cross-validate-online", new CLDACrossValOnlineCommandHandler()) );
	_command_handler_map.insert( make_pair("unrecognized", new CPureSamplingCommandHandler()) );
}

CCommandFactory::~CCommandFactory() {
}

CCommandFactory* CCommandFactory::_command_factory = 0;	// initialize the static variable to null

CCommandFactory* CCommandFactory::get_instance() {
	/////////////////////////////////////////////////
	// This block could be made thread safe        //
	if ( _command_factory == 0 ) {
		_command_factory = new CCommandFactory();
	}
	/////////////////////////////////////////////////
	return _command_factory;
}

/**
 *
 */
CCommandHandler* CCommandFactory::get_command_handler(string command_str) {
	map<string, CCommandHandler*>::iterator command_handler_map_itr;

	command_handler_map_itr = _command_handler_map.find(command_str);
	if ( command_handler_map_itr == _command_handler_map.end() ) {
		return 0;
	} else {
		return command_handler_map_itr->second;
	}
}
