/*
 * CLDACommand.cpp
 *
 *  Created on: May 24, 2010
 *      Author: pdas
 */

#include "CLDACommand.h"

using namespace std;

CLDACommand::CLDACommand() {
	// create the functionoid for LDA
	_sampling_model = new CLDASamplingFunctionoid();
}

CLDACommand::~CLDACommand() {
	delete _sampling_model;
}

/**
 * simulates the command line array of strings
 */
void CLDACommand::set_arguments(vector<string>& args) {
	_command_args = args;
}

/**
 * This function executes the command i.e. calls the execute method on the command object
 */
void CLDACommand::execute() {

	CLDAGeneralModel* lda_model = _sampling_model->get_concrete_model();
	//if (b_verbose_global) std::cout << "model is " << CCommand::_model_bytecode << endl;
	lda_model->set_bytecode(CCommand::_model_bytecode);

	size_t cmd_args_size = _command_args.size();
	if ( cmd_args_size < 2 ) {
		cout << "Cannot train or test a topic model without mentioning operation mode and the config file name" << endl;
		return;
	}

	//for ( size_t arg = 0; arg < cmd_args_size; arg++ )
	//	if (b_verbose_global) std::cout << "command argument " << arg << " = " << _command_args[arg] << " " ;
	//if (b_verbose_global) std::cout << endl;

	string model_operation_mode = _command_args[0];
	lda_model->set_model_operation_mode(model_operation_mode);	// MUST SET THIS ---- IMPORTANT!!!

	string model_config_filename = _command_args[1];

	string model_save_mode;
	if ( _command_args.size() > 2 ) 				// this is always the last item in the command parameter lists
		lda_model->set_model_dump_mode(_command_args[_command_args.size()-1] );
	else lda_model->set_model_dump_mode("text");	// default is text file dumping

	if ( model_operation_mode == "test" || model_operation_mode == "validate" )
		lda_model->reset_test_data_structures();

	// model initialization
	lda_model->read_config_file(model_config_filename);
	lda_model->read_data_from_file();

	if (!lda_model->create_and_initialize_data_structures()) {
		cout << "CLDACommand::execute() Couldn't create and initialize model data structures.. possibly wrong filepaths" << endl;
		return;
	}

	// use the functionoid for sampling
	(*_sampling_model)();

	// write model parameters to disk
	lda_model->save_model();

}
