/*
 * CLDACrossValCommand.cpp
 *
 *  Created on: Jun 23, 2010
 *      Author: pdas
 */


#include "CLDACrossValCommand.h"

using namespace std;

CLDACrossValCommand::CLDACrossValCommand() {
	_model_save_mode = "text";
	_term_to_id_dic_filename = "";

}

CLDACrossValCommand::~CLDACrossValCommand() {
	// TODO Auto-generated destructor stub
}

/**
 * simulates the command line array of strings
 */
void CLDACrossValCommand::set_arguments(vector<string>& args) {
	_command_args = args;
}

void CLDACrossValCommand::read_config_file(string model_crossval_config_filename) {
	if (b_verbose_global) std::cout << "reading from CLDACrossValCommand::read_config_file ... " << endl;

	vector<string> tokens, topic_range_token_vec;
	string delimiter = ":";

	size_t topic_num, train_perc, num_folds;
	string line, topic_range;
	ifstream fs(model_crossval_config_filename.c_str());
	if ( fs.is_open() ) {

		while ( !fs.eof() ) {
			getline( fs, line );
			if (b_verbose_global) std::cout << line << endl;

			if ( line.length() == 0 ) continue;

			tokens.clear();
			topic_range_token_vec.clear();
			///////////////////////////////////////////////////// for inputfiles
			CUtilities::tokenize(line, tokens, delimiter);
			if ( tokens.size() != 2 ) { if (b_verbose_global) std::cout << "skipping this line (" << line << ") in model config file... not in correct format.." << endl; continue; }
			CUtilities::trim(tokens[0]);
			CUtilities::trim_left_right(tokens[1]);		// the right token can be a directory

			if ( tokens[0].find("inputfile") != string::npos )
				_model_index_filename = tokens[1];
			if ( tokens[0].find("datainputmode") != string::npos )
				_data_input_mode = tokens[1];
			if ( tokens[0].find("inputdic") != string::npos )
				_term_to_id_dic_filename = tokens[1];
			if ( tokens[0].find("modelrootdir") != string::npos )
				_model_root_dir = tokens[1];
			if ( tokens[0].find("topicrange") != string::npos ) {
				topic_range = tokens[1];
				CUtilities::tokenize(topic_range, topic_range_token_vec, " " );
				for ( size_t k = 0; k < topic_range_token_vec.size(); ++k ) {
					CUtilities::string_to_number(topic_num, topic_range_token_vec[k], std::dec);
					_crossval_topics_vec.push_back( topic_num );
				}
			}
			///////////////////////////////////////////////////// for iterations
			if ( tokens[0] == "gibbsiter" ) {
				_n_gibbs_iter = tokens[1];
			}
			if ( tokens[0] == "gibbsiterationlag" ) {
				_gibbs_iter_lag = tokens[1];
			}
			if ( tokens[0] == "gibbsiterationburnin" ) {
				_gibbs_iter_burnin = tokens[1];
			}
			if ( tokens[0] == "gibbsitervalidate" || tokens[0] == "gibbsitertest" ) {
				_n_gibbs_iter_validate = tokens[1];
			}
			if ( tokens[0] == "gibbsiterationlagvalidate" || tokens[0] == "gibbsiterationlagtest" ) {
				_gibbs_iter_lag_validate = tokens[1];
			}
			if ( tokens[0] == "gibbsiterationburninvalidate" || tokens[0] == "gibbsiterationburnintest" ) {
				_gibbs_iter_burnin_validate = tokens[1];
			}
			///////////////////////////////////////////////////// for crossval
			if ( tokens[0].find("trainperc") != string::npos ) {
				CUtilities::string_to_number(train_perc, tokens[1], std::dec);
				_train_perc = train_perc;
			}
			if ( tokens[0].find("numfold") != string::npos ) {
				CUtilities::string_to_number(num_folds, tokens[1], std::dec);
				_num_folds = num_folds;
			}
			if ( tokens[0].find("modelsavemode") != string::npos ) {
				_model_save_mode = tokens[1];
			}

		}
		fs.close();
	} else {
		if (b_verbose_global) std::cout << "FATAL ERROR: Couldn't open " << model_crossval_config_filename << " .. exiting..." << endl;
		exit(APPLICATIONEXIT);
	}
}

/**
 * this functions generate the LDA config file
 * the cofig file for the model to run looks like;
 *
 * input file: /home/pdas/research/data/yahooAnswers2009/InfoXtract/index/topicmodel.input
 * num topics: 14
 * gibbs iter: 5000
 * model root dir: /home/pdas/research/data/yahooAnswers2009/InfoXtract/gibbs_lda_store/train/14T
 * gibbs iteration lag: 500
 * gibbs iteration burnin: 1000
 *
 */
string CLDACrossValCommand::generate_training_config_file_contents(size_t topic, string model_index_file, string model_root_dir) {

	string contents = "";
	contents += "input file: " + model_index_file + '\n';
	contents += "data input mode: " + _data_input_mode + '\n';
	contents += "input dict file: " + _term_to_id_dic_filename + '\n';
	contents += "num topics: "+ CUtilities::number_to_string(topic) + '\n';
	contents += "gibbs iter: "+ _n_gibbs_iter + '\n';
	contents += "model root dir: "+ model_root_dir + '\n';
	contents += "gibbs iteration lag: "+ _gibbs_iter_lag + '\n';
	contents += "gibbs iteration burnin: "+ _gibbs_iter_burnin + '\n';

	return contents;
}

/**
 * same as generate_training_config_file_contents() but iterations are modified for validation
 */
string CLDACrossValCommand::generate_validation_config_file_contents(size_t topic, string model_index_file, string model_root_dir) {

	string contents = "";
	contents += "input file: " + model_index_file + '\n';
	contents += "data input mode: " + _data_input_mode + '\n';
	contents += "input dict file: " + _term_to_id_dic_filename + '\n';
	contents += "num topics: "+ CUtilities::number_to_string(topic) + '\n';
	contents += "gibbs iter: "+ _n_gibbs_iter_validate + '\n';					// |
	contents += "model root dir: "+ model_root_dir + '\n';
	contents += "gibbs iteration lag: "+ _gibbs_iter_lag_validate + '\n';		// | these numbers can be quite different during testing or inference
	contents += "gibbs iteration burnin: "+ _gibbs_iter_burnin_validate + '\n';	// |

	return contents;
}

/**
 * This function executes the command i.e. calls the execute method on the command object
 */
void CLDACrossValCommand::execute() {

	// not needed now
	size_t cmd_args_size = _command_args.size();
	for ( size_t arg = 0; arg < cmd_args_size; arg++ )
		if (b_verbose_global) std::cout << "command argument " << arg << " = " << _command_args[arg] << " " ;
	if (b_verbose_global) std::cout << endl;

	if ( cmd_args_size < 2 ) {
		if (b_verbose_global) std::cout << "Cannot cross-validate a topic model without mentioning operation mode and the config file name" << endl;
		if (b_verbose_global) std::cout << "e.g. \"cross-validate\", \"LDA-crossval-config.txt\" " << endl;
		return;
	}

	string model_operation_mode = _command_args[0];
	string model_crossval_config_filename = _command_args[1];
	string model_save_mode;

	//////////////////////////////////////////////////////////////////
	//////////////////////////// read the cross validation config file
	read_config_file(model_crossval_config_filename);

	size_t topic;
	string training_config_filename, validation_config_filename;
	string model_index_training_filename, model_index_validation_filename;
	string train_store_root_directory, validation_store_root_dir;
	ofstream ofs_config_train, ofs_config_validate;

	for ( size_t k = 0; k < _crossval_topics_vec.size(); ++k ) {
		topic = _crossval_topics_vec[k];

		///////////////////////////////////////////////////////////////
		///////////// for each K create a new cross validation splitter
		//if (b_verbose_global) std::cout << _model_index_filename << endl << _model_root_dir << endl << _model_save_mode << endl;//---
		CTrainValidateSplitter cross_validator(
				_model_index_filename,
				_model_root_dir,
				topic,
				_num_folds,
				_model_save_mode);
		//if (b_verbose_global) std::cout << &cross_validator << endl <<  _model_index_filename << endl << _model_root_dir << endl << _train_perc << endl;//---
		///////////////////////////////////////////////////////////////
		///////////////////////////////////////////// perform splitting
		cross_validator.split();

		////////////////////////////////////////////////////////////////
		///////////////////////// generate cross validation config files
		for ( size_t f = 0; f < _num_folds; ++f ) {
			//if (b_verbose_global) std::cout << "fold # " << f << endl;
			training_config_filename = cross_validator.get_config_filenames_train_vec()[f];
			validation_config_filename = cross_validator.get_config_filenames_validation_vec()[f];

			model_index_training_filename = cross_validator.get_model_index_file_split_train_filename_vec()[f];
			model_index_validation_filename = cross_validator.get_model_index_file_split_validate_filename_vec()[f];

			train_store_root_directory = cross_validator.get_train_store_root_dir_vec()[f];
			validation_store_root_dir = cross_validator.get_validation_store_root_dir_vec()[f];

			ofs_config_train.open(training_config_filename.c_str()); ofs_config_validate.open(validation_config_filename.c_str());
			if ( ofs_config_train.is_open() && ofs_config_validate.is_open() ) {
				/////////////// write training config file
				ofs_config_train << generate_training_config_file_contents(topic, model_index_training_filename, train_store_root_directory);
				ofs_config_train.close();
				/////////////// write validation config file
				ofs_config_validate << generate_validation_config_file_contents(topic, model_index_validation_filename, validation_store_root_dir);
				ofs_config_validate << "model train root dir: " << train_store_root_directory << endl;
				ofs_config_validate.close();
			} else {
				if (b_verbose_global) std::cout << "In Cross-validation: cannot open automatically generated training config file " << training_config_filename << endl;
				if (b_verbose_global) std::cout << "                 OR: cannot open automatically generated validation config file " << validation_config_filename << endl;
				break;
			}
		}

		///////////////////////////////////////////////////////////////////
		///////////////////////////// TRAINING ////////////////////////////
		double moving_avg_likelihood = 0;	// calculate a moving average
		double moving_avg_BIC = 0;
		double moving_avg_MDL = 0;

		for ( size_t f = 0; f < _num_folds; ++f ) {
			if (b_verbose_global) std::cout << endl << "TRAINING: topic " << topic << " fold # " << f << endl;
			training_config_filename = cross_validator.get_config_filenames_train_vec()[f];
			///////////////////////////////////////////////////////////////
			////////////////////////// run cross validation for each K here

			// create the functionoid for LDACrossvalidate
			CModelSamplingFunctionoid* sampling_model = new CLDASamplingFunctionoid();
			_sampling_model = sampling_model;
			sampling_model->set_verbose_mode(true);

			CLDAGeneralModel* lda_model = sampling_model->get_concrete_model();
			lda_model->set_bytecode(CCommand::_model_bytecode);

			// some housekeeping
			lda_model->set_model_dump_mode(_model_save_mode);
			lda_model->set_model_operation_mode("train");								// MUST SET MODE ---- IMPORTANT!!!

			// model initialization
			lda_model->read_config_file(training_config_filename);						// make sure you use the training config filename here
			lda_model->read_data_from_file();											// read data from whatever is mentioned in the config file
			if ( !lda_model->create_and_initialize_data_structures() ) {				// datastructures required for model are automatically determined from model mode
				if (b_verbose_global) std::cout << "CLDACrossValCommand::execute() Couldn't create and initialize model data structures.. possibly wrong filepaths" << endl;
			} else {

				// use the functionoid for sampling
				(*sampling_model)();

				// write model parameters to disk
				lda_model->save_model();
			}

			// calculate a moving average
			if ( f == 0 ) {
				moving_avg_likelihood = sampling_model->get_final_model_log_likelihood();
				moving_avg_BIC = lda_model->get_BIC();
				moving_avg_MDL = lda_model->get_MDL();
			} else {
				moving_avg_likelihood = ((moving_avg_likelihood * (f-1+1) )+sampling_model->get_final_model_log_likelihood())/(f+1);
				moving_avg_BIC = ( moving_avg_BIC + lda_model->get_BIC() )/(f+1);
				moving_avg_MDL = ( moving_avg_MDL + lda_model->get_MDL() )/(f+1);
			}

			delete sampling_model;
		}// run model for each fold during training
		_avg_fold_train_likelihood_across_topics_vec.push_back( moving_avg_likelihood );
		_avg_fold_train_BIC_across_topics_vec.push_back(moving_avg_BIC);
		_avg_fold_train_MDL_across_topics_vec.push_back(moving_avg_MDL);

		///////////////////////////////////////////////////////////////////
		///////////////////////////// TESTING /////////////////////////////
		moving_avg_likelihood = 0;
		for ( size_t f = 0; f < _num_folds; ++f ) {
			if (b_verbose_global) std::cout << endl << "VALIDATION: topic " << topic << " fold # " << f << endl;
			validation_config_filename = cross_validator.get_config_filenames_validation_vec()[f];
			///////////////////////////////////////////////////////////////
			////////////////////////// run cross validation for each K here

			// create the functionoid for LDACrossvalidate
			CModelSamplingFunctionoid* sampling_model = new CLDASamplingFunctionoid();
			_sampling_model = sampling_model;

			sampling_model->set_verbose_mode(true);

			CLDAGeneralModel* lda_model = sampling_model->get_concrete_model();
			lda_model->set_bytecode(CCommand::_model_bytecode);

			// some housekeeping
			lda_model->set_model_dump_mode(_model_save_mode);
			lda_model->set_model_operation_mode("validate");							// MUST SET THIS!!! --IMPORTANT!

			// model initialization
			lda_model->read_config_file(validation_config_filename);					// make sure you use the validation or test config filename here
			lda_model->read_data_from_file();											// read data from whatever is mentioned in the config file
			if ( !lda_model->create_and_initialize_data_structures() ) {
				if (b_verbose_global) std::cout << "CLDACrossValCommand::execute() Couldn't create and initialize model data structures.. possibly wrong filepaths" << endl;
			} else {

				// use the functionoid for sampling
				(*sampling_model)();

				// write model parameters to disk
				lda_model->save_model();
			}

			// calculate moving averages
			if ( f == 0 ) {
				moving_avg_likelihood = sampling_model->get_final_model_log_likelihood();
				_training_avg_num_total_docs = lda_model->get_train_doc_count();
				_training_avg_vocabulary_size = lda_model->get_train_vocabulary_size();
				_training_avg_total_datum_count = lda_model->get_train_total_datum_count();
			} else {
				moving_avg_likelihood = ((moving_avg_likelihood * (f-1+1) )+sampling_model->get_final_model_log_likelihood())/(f+1);
				_training_avg_num_total_docs = ( _training_avg_num_total_docs*(f-1+1) + lda_model->get_train_doc_count())/(f+1);
				_training_avg_vocabulary_size = ( _training_avg_vocabulary_size*(f-1+1) + lda_model->get_train_vocabulary_size())/(f+1);
				_training_avg_total_datum_count = ( _training_avg_total_datum_count*(f-1+1) + lda_model->get_train_total_datum_count())/(f+1);
			}

			delete sampling_model;
		}// run model for each fold during training

		_avg_fold_validate_likelihood_across_topics_vec.push_back( moving_avg_likelihood );

	}// for each K

	compute_best_topic_setting();

	return;
}

/**
 * this function computes the best topic setting for a particular dataset
 */
void CLDACrossValCommand::compute_best_topic_setting() {
	size_t num_topics_in_range = _crossval_topics_vec.size();
	string filename = _model_root_dir + CUtilities::_file_path_separator +"fold_avg_validation_likelihoods.txt";
	FILE * fout_val_likhd = fopen(filename.c_str(), "w");
	if ( !fout_val_likhd ) {
		printf("Cannot open file %s to save!\n", filename.c_str());
		return ;
	}

	vector<double> segment_cosine_vec(num_topics_in_range, -1000);
	vector<double> validation_cosine_vec(num_topics_in_range, -1000);
	double a, b, c, d, bic, mdl;

	for ( size_t k = 1 ; k < (num_topics_in_range) ; ++k ) {
		a = (_crossval_topics_vec[k] - _crossval_topics_vec[k-1]);
		b = (-_avg_fold_validate_likelihood_across_topics_vec[k] - -_avg_fold_validate_likelihood_across_topics_vec[k-1]);
		validation_cosine_vec[k] = ( a/sqrt(a*a + b*b) );
	}

	if (b_verbose_global) std::cout << _training_avg_num_total_docs << " :: " <<  _training_avg_vocabulary_size << endl;
	fprintf ( stdout,         "  %s        %s       %s             %s             %s\n",
					"K", "TR_LKH", "VA_LKH", "BIC", "MDL");
	fprintf ( fout_val_likhd, "  %s        %s       %s             %s             %s\n",
					"K", "TR_LKH", "VA_LKH", "BIC", "MDL");
	for ( size_t k = 0 ; k < num_topics_in_range; ++k ) {
		bic = _avg_fold_train_BIC_across_topics_vec[k];
		mdl = _avg_fold_train_MDL_across_topics_vec[k];

		fprintf ( stdout,         "%3zd : %lf : %lf : %lf :  %lf\n", _crossval_topics_vec[k], _avg_fold_train_likelihood_across_topics_vec[k],
				_avg_fold_validate_likelihood_across_topics_vec[k], bic, mdl);
		fprintf ( fout_val_likhd, "%3zd : %lf : %lf : %lf : %lf\n", _crossval_topics_vec[k], _avg_fold_train_likelihood_across_topics_vec[k],
				_avg_fold_validate_likelihood_across_topics_vec[k], bic, mdl);
	}
	fprintf ( stdout, "%s\n", "--- Angle between segments ---");
	fprintf ( fout_val_likhd, "%s\n", "--- Angle between segments ---");
	/*
	 * find angle between 2 lines: ax + by = const and cx + dy = const'
	 * The x's are the topic number intervals and the y's are the negative log likelihood values for each topic number
	 */
	for ( size_t k = 1 ; k < (num_topics_in_range-1) ; ++k ) {
		a = (_crossval_topics_vec[k] - _crossval_topics_vec[k-1]);
		c = (_crossval_topics_vec[k+1] - _crossval_topics_vec[k]);
		b = (-_avg_fold_validate_likelihood_across_topics_vec[k] - -_avg_fold_validate_likelihood_across_topics_vec[k-1]);
		d = (-_avg_fold_validate_likelihood_across_topics_vec[k+1] - -_avg_fold_validate_likelihood_across_topics_vec[k]);
		segment_cosine_vec[k] = ( (b*d+a*c)/(sqrt(a*a+b*b)*sqrt(c*c+d*d)) );
	}

	fprintf ( stdout,         "  %s       %s       %s       %s       %s\n", "K", "TR_LKH", "VA_LKH", "SEGMENT-COSINE", "ANGLE");
	fprintf ( fout_val_likhd, "  %s       %s       %s       %s       %s\n", "K", "TR_LKH", "VA_LKH", "SEGMENT-COSINE", "ANGLE");
	for ( size_t k = 1 ; k < (num_topics_in_range-1) ; ++k ) {
		fprintf ( stdout,         "%3zd : %lf : %lf : %lf : %lf\n", _crossval_topics_vec[k], _avg_fold_train_likelihood_across_topics_vec[k], _avg_fold_validate_likelihood_across_topics_vec[k],
				segment_cosine_vec[k],	acos(segment_cosine_vec[k]));
		fprintf ( fout_val_likhd, "%3zd : %lf : %lf : %lf : %lf\n", _crossval_topics_vec[k], _avg_fold_train_likelihood_across_topics_vec[k], _avg_fold_validate_likelihood_across_topics_vec[k],
				segment_cosine_vec[k],	acos(segment_cosine_vec[k]));
	}
	fclose(fout_val_likhd);

}
