/*
 * CLDACrossValCommand.h
 *
 *  Created on: Jun 23, 2010
 *      Author: pdas
 */

#ifndef CLDACROSSVALCOMMAND_H_
#define CLDACROSSVALCOMMAND_H_

#include "CCommand.h"
#include "../include/includes.h"
#include "../include/CUtilities.h"
#include "../SamplingAlgorithms/CModelSamplingFunctionoid.h"
#include "../SamplingAlgorithms/CLDASamplingFunctionoid.h"
#include "../GraphicalModels/CLDAGeneralModel.h"
#include "../DataSet/CTrainValidateSplitter.h"

class CLDACrossValCommand : public CCommand {

protected:

	CModelSamplingFunctionoid* _sampling_model;

	std::vector<std::string> _command_args;
	std::vector<size_t> _crossval_topics_vec;	//<---------
	std::string _model_index_filename;			//			|
	std::string _data_input_mode;				//			|
	std::string _model_root_dir;				//			|
	std::string _n_gibbs_iter;					//			| The values from here will go into generating the individual config files
	std::string _n_gibbs_iter_validate;			//			| for training and testing
	std::string _gibbs_iter_lag;				//			| NOTE: these are all of type std::string as read from the file since there is no need for conversion
	std::string _gibbs_iter_lag_validate;		//			|
	std::string _gibbs_iter_burnin;				//			|
	std::string _gibbs_iter_burnin_validate;	//<---------


	std::string _term_to_id_dic_filename;

	size_t _train_perc;						// percentage of training data - [integer > 0]
	size_t _num_folds;						// number of folds - [integer >= 1]
	std::string _model_save_mode;				// the mode in which files are saved - either text or binary

	std::vector<double> 	_avg_fold_train_likelihood_across_topics_vec;
	std::vector<double> 	_avg_fold_validate_likelihood_across_topics_vec;
	std::vector<double> 	_avg_fold_train_BIC_across_topics_vec;
	std::vector<double> 	_avg_fold_train_MDL_across_topics_vec;

	double _training_avg_num_total_docs;
	double _training_avg_total_datum_count;
	double _training_avg_vocabulary_size;

	size_t _best_topic_setting;

public:
	CLDACrossValCommand();
	virtual ~CLDACrossValCommand();
	virtual CLDAGeneralModel* get_model() { return _sampling_model->get_concrete_model(); }

	virtual void execute();
	virtual void set_model_bytecode(std::string bytecode) { CCommand::_model_bytecode = bytecode; }
	virtual void set_arguments(std::vector<std::string>& args);
	virtual size_t get_best_topic_setting() { return _best_topic_setting; }
	virtual void compute_best_topic_setting();

protected:
	void read_config_file(std::string filename);
	std::string generate_training_config_file_contents(size_t topic, std::string model_index_file, std::string model_root_dir);
	std::string generate_validation_config_file_contents(size_t topic, std::string model_index_file, std::string model_root_dir);
};

#endif /* CLDACROSSVALCOMMAND_H_ */
