/*
 * CLDACrossValOnlineCommand.cpp
 *
 *  Created on: Jul 12, 2010
 *      Author: pdas
 */

#include "CLDACrossValCommand.h"
#include "CLDACrossValOnlineCommand.h"

using namespace std;

CLDACrossValOnlineCommand::CLDACrossValOnlineCommand() {
	CLDACrossValCommand::CLDACrossValCommand();
}

CLDACrossValOnlineCommand::~CLDACrossValOnlineCommand() {

}

void CLDACrossValOnlineCommand::execute() {
	// not needed now
	size_t cmd_args_size = _command_args.size();
	for ( size_t arg = 0; arg < cmd_args_size; arg++ )
		if (b_verbose_global) std::cout << "command argument " << arg << " = " << _command_args[arg] << " " ;
	if (b_verbose_global) std::cout << endl;

	if ( cmd_args_size < 2 ) {
		if (b_verbose_global) std::cout << "Cannot cross-validate a topic model without mentioning operation mode and the config file name" << endl;
		if (b_verbose_global) std::cout << "e.g. \"cross-validate\", \"LDA-crossval-config.txt\" " << endl;
		return;
	}

	string model_operation_mode = _command_args[0];
	string model_crossval_config_filename = _command_args[1];
	string model_save_mode;

	//////////////////////////////////////////////////////////////////
	//////////////////////////// read the cross validation config file
	read_config_file(model_crossval_config_filename);

	size_t topic;
	string training_config_filename, validation_config_filename;
	string model_index_training_filename, model_index_validation_filename;
	string train_store_root_directory, validation_store_root_dir;
	ofstream ofs_config_train, ofs_config_validate;

	for ( size_t k = 0; k < _crossval_topics_vec.size(); ++k ) {
		topic = _crossval_topics_vec[k];

		///////////////////////////////////////////////////////////////
		///////////// for each K create a new cross validation splitter
		//if (b_verbose_global) std::cout << _model_index_filename << endl << _model_root_dir << endl << _model_save_mode << endl;//---
		CTrainValidateSplitter cross_validator(
				_model_index_filename,
				_model_root_dir,
				topic,
				_num_folds,
				_model_save_mode);
		//if (b_verbose_global) std::cout << &cross_validator << endl <<  _model_index_filename << endl << _model_root_dir << endl << _train_perc << endl;//---
		///////////////////////////////////////////////////////////////
		///////////////////////////////////////////// perform splitting
		cross_validator.split();

		////////////////////////////////////////////////////////////////
		///////////////////////// generate cross validation config files
		for ( size_t f = 0; f < _num_folds; ++f ) {
			//if (b_verbose_global) std::cout << "fold # " << f << endl;
			training_config_filename = cross_validator.get_config_filenames_train_vec()[f];
			validation_config_filename = cross_validator.get_config_filenames_validation_vec()[f];

			model_index_training_filename = cross_validator.get_model_index_file_split_train_filename_vec()[f];
			model_index_validation_filename = cross_validator.get_model_index_file_split_validate_filename_vec()[f];

			train_store_root_directory = cross_validator.get_train_store_root_dir_vec()[f];
			validation_store_root_dir = cross_validator.get_validation_store_root_dir_vec()[f];

			ofs_config_train.open(training_config_filename.c_str()); ofs_config_validate.open(validation_config_filename.c_str());
			if ( ofs_config_train.is_open() && ofs_config_validate.is_open() ) {
				/////////////// write training config file
				ofs_config_train << generate_training_config_file_contents(topic, model_index_training_filename, train_store_root_directory);
				ofs_config_train.close();
				/////////////// write validation config file
				ofs_config_validate << generate_validation_config_file_contents(topic, model_index_validation_filename, validation_store_root_dir);
				ofs_config_validate << "model train root dir: " << train_store_root_directory << endl;
				ofs_config_validate.close();
			} else {
				if (b_verbose_global) std::cout << "In Cross-validation: cannot open automatically generated training config file " << training_config_filename << endl;
				if (b_verbose_global) std::cout << "                 OR: cannot open automatically generated validation config file " << validation_config_filename << endl;
				break;
			}
		}

		///////////////////////////////////////////////////////////////////
		///////////////////////////// TRAINING ////////////////////////////
		double moving_avg_likelihood = 0;	// calculate a moving average
		double moving_avg_BIC = 0;
		double moving_avg_MDL = 0;

		for ( size_t f = 0; f < _num_folds; ++f ) {
			if (b_verbose_global) std::cout << endl << "TRAINING: topic " << topic << " fold # " << f << endl;
			training_config_filename = cross_validator.get_config_filenames_train_vec()[f];
			///////////////////////////////////////////////////////////////
			////////////////////////// run cross validation for each K here

			// create the functionoid for LDACrossvalidate
			CModelSamplingFunctionoid* sampling_model = new CLDASamplingFunctionoid();
			_sampling_model = sampling_model;

			sampling_model->set_verbose_mode(true);

			CLDAGeneralModel* lda_model = sampling_model->get_concrete_model();
			lda_model->set_bytecode(CCommand::_model_bytecode);

			// some housekeeping
			lda_model->set_model_dump_mode(_model_save_mode);
			lda_model->set_model_operation_mode("train");								// MUST SET MODE ---- IMPORTANT!!!

			// model initialization
			lda_model->read_config_file(training_config_filename);						// make sure you use the training config filename here
			lda_model->read_data_from_file();											// read data from whatever is mentioned in the config file
			if ( !lda_model->create_and_initialize_data_structures() ) {				// datastructures required for model are automatically determined from model mode
				if (b_verbose_global) std::cout << "CLDACrossValCommand::execute() Couldn't create and initialize model data structures.. possibly wrong filepaths" << endl;
			} else {

				// use the functionoid for sampling
				(*sampling_model)();

				// write model parameters to disk
				lda_model->save_model();
			}

			// calculate a moving average
			if ( f == 0 ) {
				moving_avg_likelihood = sampling_model->get_final_model_log_likelihood();
				moving_avg_BIC = lda_model->get_BIC();
				moving_avg_MDL = lda_model->get_MDL();
			} else {
				moving_avg_likelihood = ((moving_avg_likelihood * (f-1+1) )+sampling_model->get_final_model_log_likelihood())/(f+1);
				moving_avg_BIC = ( moving_avg_BIC + lda_model->get_BIC() )/(f+1);
				moving_avg_MDL = ( moving_avg_MDL + lda_model->get_MDL() )/(f+1);
			}

			CLDACrossValCommand::_training_avg_num_total_docs = lda_model->get_train_doc_count();
			CLDACrossValCommand::_training_avg_vocabulary_size = lda_model->get_train_vocabulary_size();
			CLDACrossValCommand::_training_avg_total_datum_count = lda_model->get_train_total_datum_count();

			delete sampling_model;
		}// run model for each fold during training
		_avg_fold_train_likelihood_across_topics_vec.push_back( moving_avg_likelihood );
		_avg_fold_train_BIC_across_topics_vec.push_back(moving_avg_BIC);
		_avg_fold_train_MDL_across_topics_vec.push_back(moving_avg_MDL);

		/*
		 * In the online scenario, we would like to test the following:
		 * Given each fold in training, validate on all the other folds
		 * treating the topicmodel.input.validate.txt file in each of
		 * the other folds to be "sort of" unseen documents being fed in
		 * in batches to the model inference engine
		 */

		///////////////////////////////////////////////////////////////////
		///////////////////////////// TESTING /////////////////////////////
		moving_avg_likelihood = 0;
		// simulating online scenario
		size_t outer_iterations = _num_folds;
		for ( size_t f_o = 0; f_o < _num_folds/outer_iterations; ++f_o ) {

			// create the functionoid for LDACrossvalidate once for all bunches of online document sets
			CModelSamplingFunctionoid* sampling_model = new CLDASamplingFunctionoid();
			_sampling_model = sampling_model;

			for ( size_t f = 0; f < _num_folds; ++f ) {
				if (b_verbose_global) std::cout << endl << "ONLINE VALIDATION: topic " << topic << " fold # " << f << " for f_o = " << f_o << endl;
				validation_config_filename = cross_validator.get_config_filenames_validation_vec()[f];
				///////////////////////////////////////////////////////////////
				////////////////////////// run cross validation for each K here

				sampling_model->set_verbose_mode(true);

				CLDAGeneralModel* lda_model = sampling_model->get_concrete_model();
				lda_model->set_bytecode(CCommand::_model_bytecode);

				// some housekeeping
				lda_model->set_model_dump_mode(_model_save_mode);
				lda_model->set_model_operation_mode("validate");							// MUST SET THIS!!! --IMPORTANT!

				// model initialization
				lda_model->read_config_file(validation_config_filename);					// make sure you use the validation or test config filename here
				lda_model->read_data_from_file();											// read data from whatever is mentioned in the config file
				if ( !lda_model->create_and_initialize_data_structures() ) {
					if (b_verbose_global) std::cout << "CLDACrossValCommand::execute() Couldn't create and initialize model data structures.. possibly wrong filepaths" << endl;
				} else {

					// use the functionoid for sampling
					(*sampling_model)();

					// write model parameters to disk
					lda_model->save_model();
				}

				// calculate moving averages
				if ( f == 0 ) {
					moving_avg_likelihood = sampling_model->get_final_model_log_likelihood();
					_training_avg_num_total_docs = lda_model->get_train_doc_count();
					_training_avg_vocabulary_size = lda_model->get_train_vocabulary_size();
					_training_avg_total_datum_count = lda_model->get_train_total_datum_count();
				} else {
					moving_avg_likelihood = ((moving_avg_likelihood * (f-1+1) )+sampling_model->get_final_model_log_likelihood())/(f+1);
					_training_avg_num_total_docs = ( _training_avg_num_total_docs*(f-1+1) + lda_model->get_train_doc_count())/(f+1);
					_training_avg_vocabulary_size = ( _training_avg_vocabulary_size*(f-1+1) + lda_model->get_train_vocabulary_size())/(f+1);
					_training_avg_total_datum_count = ( _training_avg_total_datum_count*(f-1+1) + lda_model->get_train_total_datum_count())/(f+1);
				}

				sampling_model->get_concrete_model()->reset_test_data_structures();

			}// run model for each fold during training

			delete sampling_model;
		}// run once only as per simulation setting

		_avg_fold_validate_likelihood_across_topics_vec.push_back( moving_avg_likelihood );

	}// for each K

	CLDACrossValCommand::compute_best_topic_setting();

	return;
}
