/*
 * CLDAGibbs.cpp
 *
 *  Created on: May 21, 2010
 *      Author: pdas
 */

#include "CLDAGibbs.h"
#include "../include/cokus.h"
#include "../DataSet/CParameterMatrixReaderWriter.h"
#include "../ModelGraphBuilder/CModelGraph.h"
#include "../Printer/CPrinter.h"
#include "../include/includes.h"

using namespace std;

/**
 * @brief CLDAGibbs constructor
 * This function takes care of the initializing all the member variables to their respective default values
 */
CLDAGibbs::CLDAGibbs() {
	if (b_verbose_global) std::cout << "CLDAGIBBS OBJECT IS CREATED ONLY ONCE ****************" << endl;
	_datum_to_id_dict_filename = "";
	_model_train_saved_topic_mult_filename = "";
	_data_input_mode = "grouped";
	_accommodate_for_unknown_datum = true;

	_K = 0;

	_corpus = 0;			// initialized later on after reading the data
	_corpus_test = 0;		// initialized later on after reading the data

	_training_datum_total_count = 0;
	_V = 0;			// set initial number of words in the vocabulary to 0	|
	_D = 0;			// set the initial number of documents to 0				| very important to have these
	_V_test = 0;	// set initial number of words in the vocabulary to 0	| set in the constructor itself
	_D_test = 0;	// set the initial number of documents to 0				|

	// for 3 level vanilla LDA we have these four matrices
	_eta = 0.005;	// prior(hyperprior) parameter default value of 0.01
	_alpha = 0.005;	// prior(hyperprior) parameter default value of 0.01
	_theta = 0;		// hidden (per document topic proportion) variable - initialized later on after reading the data
	_beta = 0;		// parameter for word topic multinomials - initialized later on after reading the data

	_n_gibbs_iter = 3000;
	_gibbs_iter_lag = 50;
	_gibbs_iter_burnin = 1000;

	///////////////////////////////////////////////////////////////////////////////////////////////////////// |
	////////////////////////////////////////////////////////////// IMPORTANT - set all pointers to 0x00000000 |
	// plus we need matrices to hold counts during sampling - all initialized later on after reading the data |
	////////// for training
	_beta = 0;
	_theta = 0;
	_z_d_k = 0;			// number of times topic k is assigned to document m
	_z_v_k = 0;			// number of times topic k is assigned to a word v
	_z_d_k_total = 0;	// row-wise total of _z_d_k    - how many times topic k appears in document m
	_z_v_k_total = 0;	// column-wise total of _z_v_k - how many times word v appears in topic k
	_z_d_n = 0;			// A D x N_d matrix
	////////// for validation or testing
	_beta_test = 0;
	_theta_test = 0;
	_z_d_k_test = 0;		// number of times topic k is assigned to document m
	_z_v_k_test = 0;		// number of times topic k is assigned to a word v
	_z_d_k_total_test = 0;	// row-wise total of _z_d_k    - how many times topic k appears in document m
	_z_v_k_total_test = 0;	// column-wise total of _z_v_k - how many times word v appears in topic k
	_z_d_n_test = 0;		// A D x N_d matrix
	_z_v_k_full = 0;
	_z_v_k_total_full = 0;
	_sequential_test_vocabulary_index = 0;

}
/**
 * @brief destructor for LDAGibbs - deletes all variables to
 */
CLDAGibbs::~CLDAGibbs() {

	if ( _theta ) CTemplateMatrixMath<double>::delete_2D_matrix(_theta);	// A D x K matrix
	if ( _beta ) CTemplateMatrixMath<double>::delete_2D_matrix(_beta);		// A K x V matrix
	if ( _z_d_k ) CTemplateMatrixMath<int>::delete_2D_matrix(_z_d_k);		// A D x K matrix
	if ( _z_v_k ) CTemplateMatrixMath<int>::delete_2D_matrix(_z_v_k);		// A V x K matrix
	if ( _z_d_k_total ) delete _z_d_k_total;								// A 1 x D matrix
	if ( _z_v_k_total ) delete _z_v_k_total;								// A 1 x K matrix
	if ( _z_d_n ) CTemplateMatrixMath<int>::delete_2D_matrix(_z_d_n);		// A D x N{d} matrix

	if ( _theta_test ) CTemplateMatrixMath<double>::delete_2D_matrix(_theta_test);	// A D x K matrix
	if ( _beta_test ) CTemplateMatrixMath<double>::delete_2D_matrix(_beta_test);	// A K x V matrix
	if ( _z_d_k_test ) CTemplateMatrixMath<int>::delete_2D_matrix(_z_d_k_test);		// A D x K matrix
	if ( _z_v_k_test ) CTemplateMatrixMath<int>::delete_2D_matrix(_z_v_k_test);		// A V x K matrix
	if ( _z_d_k_total_test ) delete _z_d_k_total_test;								// A 1 x D matrix
	if ( _z_v_k_total_test ) delete _z_v_k_total_test;								// A 1 x K matrix
	if ( _z_d_n_test ) CTemplateMatrixMath<int>::delete_2D_matrix(_z_d_n_test);		// A D x N{d} matrix

	if ( _z_v_k_full ) CTemplateMatrixMath<int>::delete_2D_matrix(_z_v_k_full);
	if ( _z_v_k_total_full ) delete _z_v_k_total_full;

	if ( _corpus_test ) delete _corpus_test;
	if ( _corpus ) delete _corpus;

}

/**
 * resetting data structures for the testing phase for online scenario
 */
void CLDAGibbs::reset_test_data_structures() {
	_sequential_test_vocabulary_index = 0;
	_V_test = 0;	// set initial number of words in the vocabulary to 0	| set in the constructor itself
	_D_test = 0;	// set the initial number of documents to 0

	if ( _theta_test ) CTemplateMatrixMath<double>::delete_2D_matrix(_theta_test);	// A D x K matrix
	if ( _beta_test ) CTemplateMatrixMath<double>::delete_2D_matrix(_beta_test);	// A K x V matrix
	if ( _z_d_k_test ) CTemplateMatrixMath<int>::delete_2D_matrix(_z_d_k_test);		// A D x K matrix
	if ( _z_v_k_test ) CTemplateMatrixMath<int>::delete_2D_matrix(_z_v_k_test);		// A V x K matrix
	if ( _z_d_k_total_test ) delete _z_d_k_total_test;								// A 1 x D matrix
	if ( _z_v_k_total_test ) delete _z_v_k_total_test;								// A 1 x K matrix
	if ( _z_d_n_test ) CTemplateMatrixMath<int>::delete_2D_matrix(_z_d_n_test);		// A D x N{d} matrix

	if ( _corpus_test ) delete _corpus_test;
	_corpus_test = 0;

	_beta_test = 0;
	_theta_test = 0;
	_z_d_k_test = 0;		// number of times topic k is assigned to document m
	_z_v_k_test = 0;		// number of times topic k is assigned to a word v
	_z_d_k_total_test = 0;	// row-wise total of _z_d_k    - how many times topic k appears in document m
	_z_v_k_total_test = 0;	// column-wise total of _z_v_k - how many times word v appears in topic k
	_z_d_n_test = 0;		// A D x N_d matrix

	if ( _z_v_k ) CTemplateMatrixMath<int>::delete_2D_matrix(_z_v_k);				// A V_test x K matrix
	if ( _z_v_k_total ) delete _z_v_k_total;										// A 1 x K matrix
	_z_v_k = 0;
	_z_v_k_total = 0;

	_test_datum_id_vec.clear();							// I think this vector may be removed
	_test_datum_id_to_sequential_index_id_map.clear();	// stores the test_datum_id -> sequential_id mapping
}

/**
 * this function sets the model operation mode string here
 */
void CLDAGibbs::set_model_operation_mode(string model_operation_mode) {
	if (  model_operation_mode == "train" || model_operation_mode == "training" ||
			model_operation_mode == "est" || model_operation_mode == "estimate" ) {
		CLDAGeneralModel::_model_operation_mode = "train";
	} else if ( model_operation_mode == "test" || model_operation_mode == "testing" ||
			model_operation_mode == "validate" || model_operation_mode == "validating" || model_operation_mode == "validation" ||
			model_operation_mode == "inf" || model_operation_mode == "inference" || model_operation_mode == "inferring" ) {
		CLDAGeneralModel::_model_operation_mode = "test";
	} else {
		if (b_verbose_global) std::cout << "Couldn't recognize model operation mode = " << model_operation_mode << endl; exit(APPLICATIONEXIT);
	}
	return;
}

/**
 * Figure out the hyperparameter values here
 */
void CLDAGibbs::set_bytecode(string bytecode) {
	CLDAGeneralModel::_bytecode = bytecode; if (b_verbose_global) std::cout << endl << "In LDA Model, bytecode = " << CLDAGeneralModel::_bytecode << endl;

	CModelGraph model_graph;
	list< pair<SModelGraphSimpleNode, list<list<SModelGraphSimpleNode> > > > all_paths = model_graph.get_all_paths_from_all_obs_vars(bytecode);
	//CPrinter prn;
	//prn.print_model_simple_obsvar_paths(all_paths);
	list< pair<SModelGraphSimpleNode, list<list<SModelGraphSimpleNode> > > >::iterator ap_l_itr;

	// In basic LDA we have just one observed variable
	assert( all_paths.size() == 1 );

	SModelGraphSimpleNode& node = (*(all_paths.begin())).first;
	list<list<SModelGraphSimpleNode> >& paths = (*(all_paths.begin())).second;
	string obs_var_dim = node._dimensionality;

	// set document and word hyperparameters here
	bool doc_prior_path = false;
	list<list<SModelGraphSimpleNode> >::iterator p_itr;
	list<SModelGraphSimpleNode>::iterator s_p_itr;
	for ( p_itr = paths.begin(); p_itr != paths.end(); ++p_itr ) {
		list<SModelGraphSimpleNode>& path = *p_itr;
		// traverse along a single path
		for ( s_p_itr = path.begin(); s_p_itr != path.end(); ++s_p_itr ) {
			if ( (*s_p_itr)._dimensionality == obs_var_dim ) {
				doc_prior_path = true;
				break;
			}
		}
		if ( doc_prior_path ) {
			SModelGraphSimpleNode& hyperprior_node = path.back();
			list<string>::iterator itr;
			vector<double> param_vec;
			for ( itr = hyperprior_node._parameter_list.begin(); itr != hyperprior_node._parameter_list.end(); ++itr ) {
				double d;
				if ( CUtilities::string_to_number(d, *itr, std::dec) ) param_vec.push_back(d);
			}
			if ( param_vec.size() == 1 ) _alpha = param_vec[0];
			// else gamma(shape, scale)
		} else {
			SModelGraphSimpleNode& hyperprior_node = path.back();
			list<string>::iterator itr;
			vector<double> param_vec;
			for ( itr = hyperprior_node._parameter_list.begin(); itr != hyperprior_node._parameter_list.end(); ++itr ) {
				double d;
				if ( CUtilities::string_to_number(d, *itr, std::dec) ) param_vec.push_back(d);
			}
			if ( param_vec.size() == 1 ) _eta = param_vec[0];
			// else gamma(shape, scale);
		}
	}

	//exit(0);
}

/**
 * This function reads the configuration settings for the model
 * the configuration file looks like
 *
 */
void CLDAGibbs::read_config_file(string model_config_filename) {
	if (b_verbose_global) std::cout << "model is reading from config file " << model_config_filename << " ... "  ;

	vector<string> tokens;
	string delimiter = ":";

	int num_topics, num_gibbs_iterations, gibbs_iter_lag, gibbs_iter_burnin;

	string line;
	ifstream fs(model_config_filename.c_str());
	if ( fs.is_open() ) {

		while ( !fs.eof() ) {
			getline( fs, line );
			//if (b_verbose_global) std::cout << line << endl;	//---

			if ( line.length() == 0 ) continue;

			tokens.clear();
			CUtilities::tokenize(line, tokens, delimiter);
			if ( tokens.size() != 2 ) { if (b_verbose_global) std::cout << "skipping this line (" << line << ") in model config file... not in correct format.." << endl; continue; }
			CUtilities::trim(tokens[0]);
			CUtilities::trim_left_right(tokens[1]);

			if ( tokens[0].find("inputfile") != string::npos ) {
				_model_index_filename = tokens[1];
			}

			if ( tokens[0].find("datainputmode") != string::npos )
				_data_input_mode = tokens[1];

			if ( tokens[0].find("inputdic") != string::npos ) {
				_datum_to_id_dict_filename = tokens[1];
			}

			if ( tokens[0].find("modelrootdir") != string::npos ) {
				_model_root_dir = tokens[1];
				CUtilities::make_directory(_model_root_dir);
			}

			if ( tokens[0].find("modeltrainrootdir") != string::npos ) {
				_model_train_root_dir = tokens[1];
				CUtilities::make_directory(_model_train_root_dir);
			}

			if ( tokens[0].find("numtopics") != string::npos ) {
				CUtilities::string_to_number(num_topics, tokens[1], std::dec);
				_K = num_topics;
			}

			if ( tokens[0] == "gibbsiter" ) {
				CUtilities::string_to_number(num_gibbs_iterations, tokens[1], std::dec);
				_n_gibbs_iter = num_gibbs_iterations;
			}

			if ( tokens[0] == "gibbsiterationlag" ) {
				CUtilities::string_to_number(gibbs_iter_lag, tokens[1], std::dec);
				_gibbs_iter_lag = gibbs_iter_lag;
			}

			if ( tokens[0] == "gibbsiterationburnin" ) {
				CUtilities::string_to_number(gibbs_iter_burnin, tokens[1], std::dec);
				_gibbs_iter_burnin = gibbs_iter_burnin;
			}

		}
		fs.close();
	} else {
		if (b_verbose_global) std::cout << "FATAL ERROR: Couldn't open " << model_config_filename << " .. exiting..." << endl;
		exit(APPLICATIONEXIT);
	}
	assert(_K != 0);

	if (b_verbose_global) std::cout << "done" << endl;
}

/**
 *
 */
void CLDAGibbs::accommodate_for_unknown_datum_during_train_setup() {
	size_t n_docs = _corpus->_document_vec->size();
	cout << "accommodating for unknown words during train setup for all the " << n_docs << " documents...";

	size_t per_doc_total_unknown_datum_count;
	//double between_zero_one;
	for ( size_t d = 0; d < n_docs; ++d ) {
		// choose a random number between 4 and 6 for now
		// this random number will serve as a *total* count of no. of unknown words per doc
		//between_zero_one = (double) (((unsigned long) randomMT()) / 4294967296.0);
		//per_doc_total_unknown_datum_count = 4 + (size_t)((8*between_zero_one)/4);
		per_doc_total_unknown_datum_count = 3;	// for debugging purposes
		/*
		_corpus->_document_vec->at(d)->_datum_id_vec->push_back(_V);
		_corpus->_document_vec->at(d)->_datum_id_count_vec->push_back(per_doc_total_unknown_datum_count);
		_corpus->_document_vec->at(d)->_num_of_unique_datum = _corpus->_document_vec->at(d)->_num_of_unique_datum + 1;
		*/
		if ( _data_input_mode == "grouped" ) {
			_corpus->_document_vec->at(d)->_datum_id_vec->push_back(_V);
			_corpus->_document_vec->at(d)->_datum_id_count_vec->push_back(per_doc_total_unknown_datum_count);
			_corpus->_document_vec->at(d)->_num_of_unique_datum = _corpus->_document_vec->at(d)->_num_of_unique_datum + 1;
		}
		if ( _data_input_mode == "sequential" ) {
			for ( size_t c = 0; c < per_doc_total_unknown_datum_count; ++c ) {
				_corpus->_document_vec->at(d)->_datum_id_vec->push_back(_V);
				_corpus->_document_vec->at(d)->_datum_id_count_vec->push_back(1);
				_corpus->_document_vec->at(d)->_num_of_unique_datum = _corpus->_document_vec->at(d)->_num_of_unique_datum + 1;
			}
		}
	}
	// increase vocabulary size to accommodate for the increase in a single "UNKNOWN" = _V (previous) index
	++_V;
	cout << " done." << endl;
}

/**
 *
 */
void CLDAGibbs::read_data_from_file() {
	if ( CLDAGeneralModel::_model_dump_mode == "text" ) {
		read_data_from_text_file();								// control returns here after all the data loading
	} else if ( CLDAGeneralModel::_model_dump_mode == "binary" ) {
		read_data_from_binary_file();

	} else {
		cout << "Unsupported model read mode = " << CLDAGeneralModel::_model_dump_mode << endl;
		exit(APPLICATIONEXIT);	// if the app cannot read data, there's no point in continuing
	}
}

/**
 * read data from the index file that is in textual format
 */
void CLDAGibbs::read_data_from_text_file() {

	///////////////////////////////////////////////////////////////////////// TRAINING
	//////////////////////////////////////////////////////////////////////////////////
	if (  CLDAGeneralModel::_model_operation_mode == "train" && _data_input_mode == "grouped")
		read_grouped_data_from_text_file_for_training();
	if (  CLDAGeneralModel::_model_operation_mode == "train" && _data_input_mode == "sequential")
		read_sequential_data_from_text_file_for_training();
	if ( CLDAGeneralModel::_model_operation_mode == "train" && _accommodate_for_unknown_datum )
		accommodate_for_unknown_datum_during_train_setup();	// this is important to overcome those ugly term_id not found skips during testing

	////////////////////////////////////////////////////////////////////////// TESTING
	//////////////////////////////////////////////////////////////////////////////////
	if ( CLDAGeneralModel::_model_operation_mode == "test" && _data_input_mode == "grouped" )
		read_grouped_data_from_text_file_for_testing();
	if (  CLDAGeneralModel::_model_operation_mode == "test" && _data_input_mode == "sequential")
		read_sequential_data_from_text_file_for_testing();

}

/**
 * TRAINING
 */
void CLDAGibbs::read_grouped_data_from_text_file_for_training() {
	if ( b_verbose_global ) cout << "reading data from grouped (termID:termCount) index text file for training " << _model_index_filename << "..." << endl;

	_corpus = new SCorpus();	// create the corpus here

	vector<string> tokens;
	string delimiter = " \t\r\n";

	SDocument* doc;

	size_t num_unique_datum_per_doc;
	size_t doc_id = 0;
	size_t pos;			// to find the position of ':' in term_id:term_id_count
	size_t datum_id, datum_id_count;

	string line;
	ifstream ifs(_model_index_filename.c_str());
	if ( ifs.is_open() ) {

		while ( !ifs.eof() ) {
			getline( ifs, line );


			if ( line.length() == 0 ) continue;

			tokens.clear();
			CUtilities::tokenize(line, tokens, delimiter);

			CUtilities::string_to_number(num_unique_datum_per_doc, tokens[0], std::dec);

			/////////////////////////////////////////////////////////////////////////////
			/////////////////////////////////////////////// data is in a different format
			/////////////////////////////// the first token is the number of unique datum
			//////////////////////////////// break out and read sequential format instead
			/////////////////// This might happen if there is a mistake while indexing???
			//if(tokens.size() != num_unique_datum_per_doc + 1) {
			//	if (b_verbose_global) std::cout << "The training model index file seems to be sequential positional" << endl;
			//	is_data_sequential = true;
			//	break;
			//}
			/////////////////////////////////////////////////////////////////////////////

			doc = new SDocument();
			doc->_num_of_unique_datum = num_unique_datum_per_doc;

			for ( size_t n_d = 1; n_d <= num_unique_datum_per_doc; ++n_d ) {
				// each token is in the format datum_id:datum_id_count - no spaces allowed
				pos = tokens[n_d].find(':');
				CUtilities::string_to_number(datum_id, tokens[n_d].substr(0,pos));
				CUtilities::string_to_number(datum_id_count, tokens[n_d].substr(pos+1));
				doc->_datum_id_vec->push_back(datum_id);
				doc->_datum_id_count_vec->push_back(datum_id_count);

				_training_datum_total_count += datum_id_count;

				if ( datum_id > _V ) _V = datum_id;
			}

			_corpus->_document_vec->push_back(doc);

			doc_id++;
		}
		ifs.close();

		//if ( is_data_sequential )
		//	read_sequential_data_from_text_file_for_testing();
		//else {
			// the swap trick to trim vector capacity which a posteriori could be more than what size() returns
			vector<SDocument*>(*(_corpus->_document_vec)).swap(*(_corpus->_document_vec));

			_D = doc_id;	// number of docs in the corpus
			_V++;			// the number of datum in the voc. is the max index of a datum (assuming 0 as starting index) in the corpus + 1
			_corpus->_n_docs = _D;

		//}

	} else { cout << "Couldn't open file " << _model_index_filename << endl; exit(APPLICATIONEXIT); }
	//print_parameters();

}

/**
 * TRAINING
 */
void CLDAGibbs::read_sequential_data_from_text_file_for_training() {
	if ( b_verbose_global ) cout << "reading data from sequential (datumID) index text file for training " << _model_index_filename << "..." << endl;

	_corpus = new SCorpus();	// create the corpus here

	vector<string> tokens;
	string delimiter = " \t\r\n";

	SDocument* doc;

	size_t num_datum_per_doc;
	size_t doc_id = 0;

	map<size_t, size_t> datum_count_map;
	map<size_t, size_t>::iterator m_itr;

	string line; size_t datum_id;
	ifstream ifs(_model_index_filename.c_str());
	if ( ifs.is_open() ) {

		while ( !ifs.eof() ) {

			datum_count_map.clear();	// clear the local term count map before reading any doc
			getline( ifs, line );


			if ( line.length() == 0 ) continue;

			tokens.clear();
			CUtilities::tokenize(line, tokens, delimiter);

			CUtilities::string_to_number(num_datum_per_doc, tokens[0], std::dec);

			assert(tokens.size() == num_datum_per_doc + 1);	// the first token is the number of unique terms


			doc = new SDocument();
			for ( size_t n_d = 1; n_d <= num_datum_per_doc; ++n_d ) {
				if ( !CUtilities::string_to_number(datum_id, tokens[n_d]) ) {
					if (b_verbose_global) std::cout << "Sequential data input file is not in the right format.. reading input file in grouped format? _data_input_mode = ";
					if (b_verbose_global) std::cout << _data_input_mode << endl;
					exit(APPLICATIONEXIT);
				}

				if ( datum_id > _V ) _V = datum_id;

				doc->_datum_id_vec->push_back(datum_id);
				doc->_datum_id_count_vec->push_back(1);

				_training_datum_total_count += 1;
			}
			doc->_num_of_unique_datum = doc->_datum_id_vec->size();
			_corpus->_document_vec->push_back(doc);

			doc_id++;
		}
		ifs.close();

		// the swap trick to trim vector capacity which a posteriori could be more than what size() returns
		vector<SDocument*>(*(_corpus->_document_vec)).swap(*(_corpus->_document_vec));

		_D = doc_id;	// number of docs in the corpus
		_V++;			// the number of terms in the voc. is the max index of a term (assuming 0 as starting index) in the corpus + 1
		_corpus->_n_docs = _D;

	} else { cout << "Couldn't open file " << _model_index_filename << endl; exit(APPLICATIONEXIT); }

}

/**
 * TESTING
 */
void CLDAGibbs::read_grouped_data_from_text_file_for_testing() {

	if ( b_verbose_global ) cout << "reading data from grouped (termID:termCount) index text file for validation or testing " << _model_index_filename << "..." << endl;

	string extension = ".txt";

	// the model.other file will contain all one dimensional parameters i.e. the scalars
	string model_other_filename = _model_train_root_dir + CUtilities::_file_path_separator + "model.other.train" + extension;

	// other
	FILE * fin_model_other = fopen(model_other_filename.c_str(), "r");
	if ( !fin_model_other ) {
		printf("Cannot open file %s to load!\n", model_other_filename.c_str());
		return ;
	}

	/////////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////////////////////// write the symmetric dirichlet priors here
	//////////////////////////////////////// using fscanf here, we can get automatic tokenization
	int num_of_items_read;
	num_of_items_read = fscanf( fin_model_other, "alpha: %lf\n", &_alpha);
	num_of_items_read = fscanf( fin_model_other, "eta: %lf\n", &_eta);		// reading this file once more is much efficient
	num_of_items_read = fscanf( fin_model_other, "num docs: %zd\n", &_D);	// that recompacting the term_id vectors
	num_of_items_read = fscanf( fin_model_other, "num words: %zd\n", &_V);	// in _corpus_test based on _V in training
	num_of_items_read = fscanf( fin_model_other, "num topics: %zd\n", &_K);
	if ( fin_model_other ) fclose( fin_model_other );

	/////////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////////////////////////////// reading the actual test data here

	_corpus_test = new SCorpus();	// create the corpus here
	_sequential_test_vocabulary_index = 0;
	map<size_t, size_t>::iterator stvi_map_itr;

	assert(_test_datum_id_to_sequential_index_id_map.size() == 0);	// just to be sure

	vector<string> tokens;
	string delimiter = " \t\r\n";

	SDocument* doc;

	size_t num_unique_datum_per_doc;
	size_t doc_id = 0;
	size_t pos;			// to find the position of ':' in term_id:term_id_count
	size_t datum_id, datum_id_count;

	string line;
	ifstream ifs_test(_model_index_filename.c_str());
	if ( ifs_test.is_open() ) {

		while ( !ifs_test.eof() ) {
			getline( ifs_test, line );


			if ( line.length() == 0 ) continue;

			tokens.clear();
			CUtilities::tokenize(line, tokens, delimiter);

			CUtilities::string_to_number(num_unique_datum_per_doc, tokens[0], std::dec);

			/////////////////////////////////////////////////////////////////////////////
			/////////////////////////////////////////////// data is in a different format
			/////////////////////////////// the first token is the number of unique terms
			//////////////////////////////// break out and read sequential format instead
			/////////////////// This might happen if there is a mistake while indexing???
			//if(tokens.size() != num_unique_datum_per_doc + 1) {
			//	if (b_verbose_global) std::cout << "The test or validation model index file seems to be sequential positional" << endl;
			//	is_data_sequential = true;
			//	break;
			//}
			//////////////////////////////////////////////////////////////////////////////

			doc = new SDocument();

			size_t unseen_word_term_count = 0;

			for ( size_t n_d = 1; n_d <= num_unique_datum_per_doc; ++n_d ) {
				// each token is in the format term_id:term_id_count - no spaces allowed
				pos = tokens[n_d].find(':');
				CUtilities::string_to_number(datum_id, tokens[n_d].substr(0,pos));
				CUtilities::string_to_number(datum_id_count, tokens[n_d].substr(pos+1));

				// signal for unseen term - not really needed for testing to find optimum _K
				if ( (!_accommodate_for_unknown_datum) && (datum_id >= _V) ) {
					if (b_verbose_global) std::cout << "Unseen term_id " << datum_id << " is skipped" << endl;
					continue;
				}
				// accomodate for unseen terms
				if ( (_accommodate_for_unknown_datum) && (datum_id > (_V - 2)) ) {
					if (b_verbose_global) std::cout << "Unseen term_id " << datum_id << " found. " ;
					if (b_verbose_global) std::cout << "Clubbing all unknown term_ids into one single term_id: " << _V-1 << endl;
					datum_id = _V - 1;
					++unseen_word_term_count;
				}

				if ( datum_id < (_V-1) ) {
					doc->_datum_id_vec->push_back(datum_id);
					doc->_datum_id_count_vec->push_back(datum_id_count);
				}

				// if the test term_id has never been read before, just add it to the set for reference
				stvi_map_itr = _test_datum_id_to_sequential_index_id_map.find(datum_id);
				if ( stvi_map_itr == _test_datum_id_to_sequential_index_id_map.end() ) {
					_test_datum_id_to_sequential_index_id_map.insert( make_pair(datum_id, _sequential_test_vocabulary_index) );
					++_sequential_test_vocabulary_index;
				}

				if ( datum_id > _V_test ) _V_test = datum_id;	// initial setting of _V_test here... will be changed during loading of the training model
															// into a sparse compact representation of a vector of term_ids
			}

			if ( unseen_word_term_count > 0 ) {
				doc->_datum_id_vec->push_back(_V-1);
				doc->_datum_id_count_vec->push_back(unseen_word_term_count);
			}

			doc->_num_of_unique_datum = doc->_datum_id_vec->size();
			_corpus_test->_document_vec->push_back(doc);

			doc_id++;
		}
		ifs_test.close();

		//if ( is_data_sequential )
		//	read_sequential_data_from_text_file_for_testing();
		//else {
			// the swap trick to trim vector capacity which a posteriori could be more than what size() returns
			vector<SDocument*>(*(_corpus_test->_document_vec)).swap(*(_corpus_test->_document_vec));

			_D_test = doc_id;	// number of docs in the corpus
			_V_test = _test_datum_id_to_sequential_index_id_map.size();	// this is the total number of words seen during reading of the test file (NOT THE MAX TERM ID)
			_corpus_test->_n_docs = _D_test;	// important set this or else destructor will segment fault

			// populate vector<size_t> _test_datum_id_vec here
			assert(_test_datum_id_vec.size() == 0);							// just to be sure
			_test_datum_id_vec.resize(_test_datum_id_to_sequential_index_id_map.size());
			for ( stvi_map_itr = _test_datum_id_to_sequential_index_id_map.begin();
					stvi_map_itr != _test_datum_id_to_sequential_index_id_map.end(); ++stvi_map_itr)
				_test_datum_id_vec.at(stvi_map_itr->second) = (stvi_map_itr->first);	// pushing in the actual test term_ids
		//}

	} else { cout << "Couldn't open file " << _model_index_filename << endl; exit(APPLICATIONEXIT); }

	// finally - after all the data loading and the creation of the _test_datum_id_to_sequential_index_id_map
	if ( !load_model() ) {
		cout << "couldn't load training model parameters.. exiting" << endl; exit(APPLICATIONEXIT);
	}

}

/**
 * TESTING
 */
void CLDAGibbs::read_sequential_data_from_text_file_for_testing() {

	if ( b_verbose_global ) cout << "reading data from sequential (termID) index text file for validation or testing " << _model_index_filename << "..." << endl;

	string extension = ".txt";

	// the model.other file will contain all one dimensional parameters i.e. the scalars
	string model_other_filename = _model_train_root_dir + CUtilities::_file_path_separator + "model.other.train" + extension;

	// other
	FILE * fin_model_other = fopen(model_other_filename.c_str(), "r");
	if ( !fin_model_other ) {
		printf("Cannot open file %s to load!\n", model_other_filename.c_str());
		return ;
	}

	/////////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////////////////////// write the symmetric dirichlet priors here
	//////////////////////////////////////// using fscanf here, we can get automatic tokenization
	int num_of_items_read;
	num_of_items_read = fscanf( fin_model_other, "alpha: %lf\n", &_alpha);
	num_of_items_read = fscanf( fin_model_other, "eta: %lf\n", &_eta);		// reading this file once more is much efficient
	num_of_items_read = fscanf( fin_model_other, "num docs: %zd\n", &_D);	// that recompacting the term_id vectors
	num_of_items_read = fscanf( fin_model_other, "num words: %zd\n", &_V);	// in _corpus_test based on _V in training
	num_of_items_read = fscanf( fin_model_other, "num topics: %zd\n", &_K);
	if ( fin_model_other ) fclose( fin_model_other );

	/////////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////////////////////////////// reading the actual test data here

	_corpus_test = new SCorpus();	// create the corpus here
	_sequential_test_vocabulary_index = 0;
	map<size_t, size_t>::iterator stvi_map_itr;

	vector<string> tokens;
	string delimiter = " \t\r\n";

	SDocument* doc;

	size_t num_datum_per_doc;
	size_t doc_id = 0;

	map<size_t, size_t> datum_count_map;
	map<size_t, size_t>::iterator m_itr;

	string line; size_t datum_id;
	ifstream ifs_test(_model_index_filename.c_str());
	if ( ifs_test.is_open() ) {

		while ( !ifs_test.eof() ) {

			datum_count_map.clear();	// clear the local term count map before reading any doc
			getline( ifs_test, line );


			if ( line.length() == 0 ) continue;

			tokens.clear();
			CUtilities::tokenize(line, tokens, delimiter);

			CUtilities::string_to_number(num_datum_per_doc, tokens[0], std::dec);

			assert(tokens.size() == num_datum_per_doc + 1);	// the first token is the number of unique terms

			// now process the term_count_map
			doc = new SDocument();

			for ( size_t n_d = 1; n_d <= num_datum_per_doc; ++n_d ) {

				if ( !CUtilities::string_to_number(datum_id, tokens[n_d]) ) {
					if (b_verbose_global) std::cout << "Sequential data input file is not in the right format.. reading input file in grouped format? _data_input_mode = ";
					if (b_verbose_global) std::cout << _data_input_mode << endl;
					exit(APPLICATIONEXIT);
				}
				//if (b_verbose_global) std::cout << term_id << " ";
				// signal for unseen term - not really needed for testing to find optimum _K
				if ( (!_accommodate_for_unknown_datum) && (datum_id >= _V) ) {
					if (b_verbose_global) std::cout << "Unseen term_id " << datum_id << " is skipped" << endl;
					continue;
				}
				// accomodate for unseen terms
				if ( (_accommodate_for_unknown_datum) && (datum_id > (_V - 2)) ) {
					if (b_verbose_global) std::cout << "Unseen term_id " << datum_id << " found. " ;
					if (b_verbose_global) std::cout << "Clubbing all unknown term_ids into one single term_id: " << _V-1 << endl;
					datum_id = _V - 1;
				}
				doc->_datum_id_vec->push_back( datum_id );
				doc->_datum_id_count_vec->push_back( 1 );

				// if the test term_id has never been read before, just add it to the set for reference
				stvi_map_itr = _test_datum_id_to_sequential_index_id_map.find(datum_id);
				if ( stvi_map_itr == _test_datum_id_to_sequential_index_id_map.end() ) {
					_test_datum_id_to_sequential_index_id_map.insert( make_pair(datum_id, _sequential_test_vocabulary_index) );
					++_sequential_test_vocabulary_index;
				}
			}

			doc->_num_of_unique_datum = doc->_datum_id_vec->size();
			_corpus_test->_document_vec->push_back(doc);

			doc_id++;
		}
		ifs_test.close();

		// the swap trick to trim vector capacity which a posteriori could be more than what size() returns
		vector<SDocument*>(*(_corpus_test->_document_vec)).swap(*(_corpus_test->_document_vec));

		_D_test = doc_id;											// number of docs in the corpus
		_V_test = _test_datum_id_to_sequential_index_id_map.size();	// this is the total number of words seen during reading of the test file (NOT THE MAX TERM ID)
		_corpus_test->_n_docs = _D_test;

		// populate vector<size_t> _test_datum_id_vec here
		assert(_test_datum_id_vec.size() == 0);							// just to be sure
		_test_datum_id_vec.resize(_test_datum_id_to_sequential_index_id_map.size());
		for ( stvi_map_itr = _test_datum_id_to_sequential_index_id_map.begin();
				stvi_map_itr != _test_datum_id_to_sequential_index_id_map.end(); ++stvi_map_itr)
			_test_datum_id_vec.at(stvi_map_itr->second) = (stvi_map_itr->first);	// pushing in the actual test term_ids

	} else { cout << "Couldn't open file " << _model_index_filename << endl; exit(APPLICATIONEXIT); }

	// finally - after all the data loading and the creation of the _test_datum_id_to_sequential_index_id_map
	if ( !load_model() ) {
			cout << "couldn't load training model parameters.. exiting" << endl; exit(APPLICATIONEXIT);
	}
}


/**
 * read data from the index file that is binary
 */
void CLDAGibbs::read_data_from_binary_file() {
	// TODO: to be implemented
}

/**
 * TRAINING
 */
void CLDAGibbs::read_grouped_data_from_binary_file_for_training() {
	// TODO: to be implemented
}
/**
 * TRAINING
 */
void CLDAGibbs::read_grouped_data_from_binary_file_for_testing() {
	// TODO: to be implemented
}
/**
 * TESTING
 */
void CLDAGibbs::read_sequential_data_from_binary_file_for_training() {
	// TODO: to be implemented
}
/**
 * TESTING
 */
void CLDAGibbs::read_sequential_data_from_binary_file_for_testing() {
	// TODO: to be implemented
}


/**
 * some printing functions for diagnostics
 */
void CLDAGibbs::print_parameters() {
	if (b_verbose_global) std::cout << "\nPRINTING MODEL SETTINGS............................... " << endl;
	if (b_verbose_global) std::cout << "input file = " << _model_index_filename << endl;
	if (b_verbose_global) std::cout << "model root = " << _model_root_dir << endl;
	if (b_verbose_global) std::cout << "num topics = " << _K << endl;
	if (b_verbose_global) std::cout << "num iterations = " << _n_gibbs_iter << endl;
	if (b_verbose_global) std::cout << "burnin at " << _gibbs_iter_burnin << endl;
	if (b_verbose_global) std::cout << "lag at " << _gibbs_iter_lag << endl;

	if (b_verbose_global) std::cout << "_eta = " << _eta << endl;
	if (b_verbose_global) std::cout << "_alpha = " << _alpha << endl;

	////////////////////////////////////////////////
	// diagnostics - printing corpus sizes
	//for ( doc_id = 0; doc_id < _corpus->_document_vec->size(); doc_id++ )
	//	if (b_verbose_global) std::cout << _corpus->_document_vec->at(doc_id)->_num_of_unique_datum << ":" << _corpus->_document_vec->at(doc_id)->_datum_id_vec->size() << " " ;
	//if (b_verbose_global) std::cout << endl;
	if ( _z_d_k ) if (b_verbose_global) std::cout << "_zdk dimensions = " << _z_d_k->size() << "x" << _z_d_k->at(0)->size() << endl;
	if ( _z_v_k ) if (b_verbose_global) std::cout << "_zvk dimensions = " << _z_v_k->size() << "x" << _z_v_k->at(0)->size() << endl;
	if ( _z_d_k_total ) if (b_verbose_global) std::cout << "_zdk_total dimensions = " << _z_d_k_total->size() << endl;
	if ( _z_v_k_total ) if (b_verbose_global) std::cout << "_zvk_total dimensions = " << _z_v_k_total->size() << endl;
	if ( _theta ) if (b_verbose_global) std::cout << "_theta = " << _theta->size() << "x" << _theta->at(0)->size() << endl;
	if ( _beta ) if (b_verbose_global) std::cout << "_beta = " << _beta->size() << "x" << _beta->at(0)->size() << endl;
	if (b_verbose_global) std::cout << "docs = " << _D << " max_train_vocabulary_index + 1 = " << _V << " num topics = " << _K << endl;
	if (b_verbose_global) std::cout << "total count of training datum = " << _training_datum_total_count << endl;

	if (b_verbose_global) std::cout << "........................................................ WHILE TRAINING" << endl;

	if ( _model_operation_mode == "test" ) {
		if (b_verbose_global) std::cout << "input file = " << _model_index_filename << endl;
		if (b_verbose_global) std::cout << "model root = " << _model_root_dir << endl;
		if (b_verbose_global) std::cout << "num topics = " << _K << endl;
		if (b_verbose_global) std::cout << "num iterations = " << _n_gibbs_iter << endl;
		if (b_verbose_global) std::cout << "burnin at " << _gibbs_iter_burnin << endl;
		if (b_verbose_global) std::cout << "lag at " << _gibbs_iter_lag << endl;

		if (b_verbose_global) std::cout << "_eta = " << _eta << endl;
		if (b_verbose_global) std::cout << "_alpha = " << _alpha << endl;

		if ( _z_d_k_test ) if (b_verbose_global) std::cout << "_zdk_test dimensions = " << _z_d_k_test->size() << "x" << _z_d_k_test->at(0)->size() << endl;
		if ( _z_v_k_test ) if (b_verbose_global) std::cout << "_zvk_test dimensions = " << _z_v_k_test->size() << "x" << _z_v_k_test->at(0)->size() << endl;
		if ( _z_d_k_total_test ) if (b_verbose_global) std::cout << "_zdk_total_test dimensions = " << _z_d_k_total_test->size() << endl;
		if ( _z_v_k_total_test ) if (b_verbose_global) std::cout << "_zvk_total_test dimensions = " << _z_v_k_total_test->size() << endl;
		if ( _theta_test ) if (b_verbose_global) std::cout << "_theta_test = " << _theta_test->size() << "x" << _theta_test->at(0)->size() << endl;
		if ( _beta_test ) if (b_verbose_global) std::cout << "_beta_test = " << _beta_test->size() << "x" << _beta_test->at(0)->size() << endl;
		if (b_verbose_global) std::cout << "docs = " << _D_test << " max_sequential_test_vocabulary_index + 1 = " << _V_test << " num topics = " << _K << endl;
	}

	if (b_verbose_global) std::cout << "......................................................... WHILE TESTING" << endl;

	////////////////////////////////////////////////
}

/**
 * Load model from files written in chosen mode
 */
bool CLDAGibbs::load_model() {
	if ( CLDAGeneralModel::_model_dump_mode == "text" )
		return load_model_in_text_mode();
	else if ( CLDAGeneralModel::_model_dump_mode == "binary" )
		return load_model_in_binary_mode();
	return false;
}

/**
 *
 * This function loads already saved model files in ascii text format from a specified model_root_directory:
 *
 *  double _eta;						// ( symmetric ) dirichlet prior for beta (topics themselves)	| saved in model.other
 *  double _alpha;						// ( symmetric ) dirichlet prior for theta (topic proportions)	| saved in model.other
 *	vector<vector<double>* >* _theta;	// A D x K matrix												| saved in model.theta
 *	vector<vector<double>* >* _beta;	// A K x V matrix												| saved in model.beta
 *
 *	vector<vector<int>* >* _z_d_k;		// number of times topic k is assigned to document m
 *	vector<vector<int>* >* _z_v_k;		// number of times topic k is assigned to a word v
 *	vector<int>*		   _z_d_k_total;// row-wise total of _z_d_k    - how many times topic k appears in document m
 *	vector<int>*		   _z_v_k_total;// column-wise total of _z_v_k - how many times word v appears in topic k
 *	vector<vector<int>* >* _z_d_n;		// A D x N_d matrix
 *
 *	string _model_operation_mode;				// string to specify whether the model is "trained" or being "tested"
 *	string _model_index_filename;		// full path for the model index file
 *	string _model_root_dir;				// full path of the directory to store the output files
 *	size_t _K;							// number of topics												| saved in model.other
 *	size_t _V;							// number of terms in the vocabulary							| saved in model.other
 *	size_t _D;							// number of docs in the corpus									| saved in model.other
 *	size_t _n_gibbs_iter;				// number of gibbs iterations
 *
 *	double _log_prob;					// likelihood under this model for a given value of K			| saved in model.other
 *
 * NOTE: model_root_dir is the absolute path to the directory where all the training model files are saved
 * 		 _test_datum_id_to_sequential_index_id_map is created during reading of the test data - maps test_term_id -> [0, _V_test]
 *
 */
bool CLDAGibbs::load_model_in_text_mode() {

	string model_root_dir = "";
	if ( _model_operation_mode == "train" ) model_root_dir = _model_root_dir;
	if ( _model_operation_mode == "test" ) model_root_dir = _model_train_root_dir;

	if ( b_verbose_global ) cout << "loading model files from " << model_root_dir << " ..." ;

	bool load_status = false;

	string extension = ".txt";

	// the model.other file will contain all one dimensional parameters i.e. the scalars
	string model_other_filename = model_root_dir + CUtilities::_file_path_separator + "model.other.train" + extension;

	// the model.theta file will contain the document topic proportion matrix
	string model_theta_filename = model_root_dir + CUtilities::_file_path_separator + "model.theta.train" + extension;

	// the model.beta file will contain the word topic distribution matrix
	string model_beta_filename = model_root_dir + CUtilities::_file_path_separator + "model.beta.train" + extension;
	_model_train_saved_topic_mult_filename = model_beta_filename;

	// the model.wordassign file will contain the word topic distribution matrix
	string model_word_assign_filename = model_root_dir + CUtilities::_file_path_separator + "model.wordtopicassign.train" + extension;

	// other
	FILE * fin_model_other = fopen(model_other_filename.c_str(), "r");
	if ( !fin_model_other ) {
		printf("Cannot open file %s to load!\n", model_other_filename.c_str());
		return load_status;
	}
	// word assign - let this be a stream - easier to tokenize
	ifstream ifs_model_word_assign(model_word_assign_filename.c_str());
	if ( !ifs_model_word_assign.is_open() ) {
		printf("Cannot open file %s to load!\n", model_word_assign_filename.c_str());
		return load_status;
	}

	/////////////////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////// read the symmetric dirichlet priors here
	//////////////////////////////////////// using fscanf here, we can get automatic tokenization
	int num_of_items_read;
	num_of_items_read = fscanf( fin_model_other, "alpha: %lf\n", &_alpha);
	num_of_items_read = fscanf( fin_model_other, "eta: %lf\n", &_eta);
	num_of_items_read = fscanf( fin_model_other, "num docs: %zd\n", &_D);		// no harm done in reading these
	num_of_items_read = fscanf( fin_model_other, "num words: %zd\n", &_V);		// things again
	num_of_items_read = fscanf( fin_model_other, "num words total: %ld\n", &_training_datum_total_count);
	num_of_items_read = fscanf( fin_model_other, "num topics: %zd\n", &_K);
	if ( fin_model_other ) fclose( fin_model_other );

	/////////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////// read model assign here

	////////////////////////////// note //////////////
	// the model assign file doesn't have term counts
	// instead each term_id:topic_id pair is written
	// for however times term_id occurs in the document
	// see function save_model_to_[text/binary]_file()

	if ( _z_v_k_full == 0 && _z_v_k_total_full == 0 ) {	// for online scenario, load _z_v_k_full and _z_v_k_total_full only once
		//////// ----- ////////
		CTemplateMatrixMath<int>::create_2D_matrix(&_z_v_k_full, _V, _K);	// _V is properly set here since it is called from read_data() for test
		_z_v_k_total_full = new vector<int>(_K, 0);							// initialize with 0 -- important

		size_t pos_colon_separator;
		size_t doc_id, training_datum_id, training_datum_topic;
		string line;
		vector<string> token_vec;

		doc_id = 0;
		while ( !ifs_model_word_assign.eof() ) {
			getline( ifs_model_word_assign, line );
			CUtilities::trim_left_right(line);
			if ( line.length() == 0 ) continue;

			// a document is grabbed at this point and it looks like
			// term_id:topic_id ... (repeated for ALL POSITIONS in the document)
			token_vec.clear();
			CUtilities::tokenize(line, token_vec, " \n\r\t");

			size_t num_tokens = token_vec.size();
			for ( size_t n_d = 0; n_d < num_tokens; ++n_d) {
				// each token is in the format term_id:term_id_count - no spaces allowed
				pos_colon_separator = token_vec[n_d].find(':');
				CUtilities::string_to_number(training_datum_id, token_vec[n_d].substr(0,pos_colon_separator));
				CUtilities::string_to_number(training_datum_topic, token_vec[n_d].substr(pos_colon_separator+1));
				(*(*_z_v_k_full)[training_datum_id])[training_datum_topic] += 1;
				(*_z_v_k_total_full)[training_datum_topic] += 1;

			}


			++doc_id;
		}
		// closing descriptors for custom files
		ifs_model_word_assign.close();
		//////// ----- ////////
	} else {
		if (b_verbose_global) std::cout << "_z_v_k_full != 0 && _z_v_k_total_full != 0 and of sizes: " << _z_v_k_full->size() << " and " << _z_v_k_total_full->size() << endl;
	}

	assert( _z_v_k == 0 && _z_v_k_total == 0 );
	// we need one more to retain the counts of topics assigned each matching word in the training corpus a.k.a. _z_v_k
	CTemplateMatrixMath<int>::create_2D_matrix(&_z_v_k, _V_test, _K);	// _V_test is properly set here since it is called from read_data() for test
	_z_v_k_total = new vector<int>(_K, 0);								// initialize with 0 -- important

	assert(_test_datum_id_vec.size() == _V_test);

	// create the smaller _z_v_k of dimensionality _V_test x K matrix for faster iterations over test data
	for ( size_t seq_v_test = 0; seq_v_test < _V_test; ++seq_v_test ) {
		for ( size_t k = 0; k < _K; ++k ) {
			(*(*_z_v_k)[ seq_v_test ])[k]
			                    = (*(*_z_v_k_full)[ _test_datum_id_vec[seq_v_test] ])[k];	// _test_datum_id_to_sequential_index_id_map[ _test_datum_id_vec[seq_v_test] ]
			(*_z_v_k_total)[k] += (*(*_z_v_k_full)[ _test_datum_id_vec[seq_v_test] ])[k];
		}
	}


	load_status = true;
	if ( b_verbose_global ) cout << "done." << endl;
	return load_status;
}

/**
 * loads already saved model files from a specified model_root_directory
 */
bool CLDAGibbs::load_model_in_binary_mode() {
	// TODO: to be implemented
	bool load_status = false;

	return load_status;
}

/**
 * Save model in chosen mode
 */
void CLDAGibbs::save_model() {
	if ( (CLDAGeneralModel::_model_dump_mode == "text") &&  (CLDAGeneralModel::_model_operation_mode == "train") )
		save_training_model_in_text_mode();
	if ( (CLDAGeneralModel::_model_dump_mode == "text") && (CLDAGeneralModel::_model_operation_mode == "test") )
		save_test_model_in_text_mode();

	if ( (CLDAGeneralModel::_model_dump_mode == "binary") && (CLDAGeneralModel::_model_operation_mode == "train") )
		save_training_model_in_binary_mode();
	if ( (CLDAGeneralModel::_model_dump_mode == "binary") && (CLDAGeneralModel::_model_operation_mode == "test") )
		save_test_model_in_binary_mode();
}

/**
 * This function saves the following data structures of the model in a readable text file:
 *
 * _model_root_dir is the absolute path to the directory where all the model files are saved
 */
void CLDAGibbs::save_training_model_in_text_mode() {
	if ( b_verbose_global ) cout << "saving model files in " << _model_root_dir << " ..." ;
	string extension = ".txt";

	// the model.other file will contain all one dimensional parameters i.e. the scalars
	string model_other_filename = _model_root_dir + CUtilities::_file_path_separator + "model.other.train" + extension;

	// the model.theta file will contain the document topic proportion matrix
	string model_theta_filename = _model_root_dir + CUtilities::_file_path_separator + "model.theta.train" + extension;
	_model_train_saved_topic_mult_filename = _model_root_dir + CUtilities::_file_path_separator + "model.theta.train" + extension;

	// the model.beta file will contain the word topic distribution matrix
	string model_beta_filename = _model_root_dir + CUtilities::_file_path_separator + "model.beta.train" + extension;

	// the model.wordassign file will contain the word topic distribution matrix
	string model_word_assign_filename = _model_root_dir + CUtilities::_file_path_separator + "model.wordtopicassign.train" + extension;

	// other
	FILE * fout_model_other = fopen(model_other_filename.c_str(), "w");
	if ( !fout_model_other ) {
		printf("Cannot open file %s to save!\n", model_other_filename.c_str());
		return ;
	}
	// word assign
	FILE * fout_model_word_assign = fopen(model_word_assign_filename.c_str(), "w");
	if ( !fout_model_word_assign ) {
		printf("Cannot open file %s to save!\n", model_word_assign_filename.c_str());
		return ;
	}

	/////////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////////////////////// write the symmetric dirichlet priors here
	fprintf( fout_model_other, "alpha: %10.6f\n", _alpha);
	fprintf( fout_model_other, "eta: %10.6f\n", _eta);
	fprintf( fout_model_other, "num docs: %zd\n", _D);
	fprintf( fout_model_other, "num words: %zd\n", _V);
	fprintf( fout_model_other, "num words total: %ld\n", _training_datum_total_count);
	fprintf( fout_model_other, "num topics: %zd\n", _K);

	/////////////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////////////////////////////////////////////// write model assign here
	size_t n_datum, datum_count;
	for ( size_t d = 0; d < _D; ++d ) {
		vector<int> *term_id_vec = _corpus->_document_vec->at(d)->_datum_id_vec;
		vector<int> *term_id_count_vec = _corpus->_document_vec->at(d)->_datum_id_count_vec;
		n_datum = term_id_vec->size();
		for ( size_t n = 0; n < n_datum; ++n ) {
			datum_count = term_id_count_vec->at(n);
			for ( size_t c = 0; c < datum_count; ++c )
				fprintf( fout_model_word_assign, " %d:%d", term_id_vec->at(n), (*(*_z_d_n)[d])[n] );
		}
		fprintf( fout_model_word_assign, "\n"); fflush(fout_model_word_assign);	// flushed and written a full document
	}
	// closing descriptors for custom files
	if ( fout_model_other ) fclose( fout_model_other );
	if ( fout_model_word_assign ) fclose( fout_model_word_assign );

	/////////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////// write parameters and hidden variables here
	CParameterMatrixReaderWriter pmrw;
	// theta
	pmrw.write_parameter_matrix_in_text_mode(model_theta_filename, _theta);
	// beta
	pmrw.write_parameter_matrix_in_text_mode(model_beta_filename, _beta);
	cout << "done." << endl;
}

/**
 *
 */
void CLDAGibbs::save_test_model_in_text_mode() {
	if ( b_verbose_global ) cout << "saving model files in " << _model_root_dir << "... " ;
	string extension = ".txt";

	// the model.other file will contain all one dimensional parameters i.e. the scalars
	string model_other_filename = _model_root_dir + CUtilities::_file_path_separator + "model.other.test" + extension;

	// the model.other file will contain all one dimensional parameters i.e. the scalars
	string model_reducedvocab_filename = _model_root_dir + CUtilities::_file_path_separator + "model.reducedvocab.test.index" + extension;

	// the model.theta file will contain the document topic proportion matrix
	string model_theta_filename = _model_root_dir + CUtilities::_file_path_separator + "model.theta.test" + extension;

	// the model.beta file will contain the word topic distribution matrix
	string model_beta_filename = _model_root_dir + CUtilities::_file_path_separator + "model.beta.test" + extension;

	// the model.wordassign file will contain the word topic distribution matrix
	string model_word_assign_filename = _model_root_dir + CUtilities::_file_path_separator + "model.wordtopicassign.test" + extension;

	// other
	FILE * fout_model_other = fopen(model_other_filename.c_str(), "w");
	if ( !fout_model_other ) {
		printf("Cannot open file %s to save!\n", model_other_filename.c_str());
		return ;
	}
	// word assign
	FILE * fout_model_word_assign = fopen(model_word_assign_filename.c_str(), "w");
	if ( !fout_model_word_assign ) {
		printf("Cannot open file %s to save!\n", model_word_assign_filename.c_str());
		return ;
	}
	// reduced vocab
	FILE * fout_reducedvocab_other = fopen(model_reducedvocab_filename.c_str(), "w");
	if ( !fout_reducedvocab_other ) {
		printf("Cannot open file %s to save!\n", model_reducedvocab_filename.c_str());
		return ;
	}

	/////////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////// write the scalars here
	fprintf( fout_model_other, "alpha: %10.6f\n", _alpha);
	fprintf( fout_model_other, "eta: %10.6f\n", _eta);
	fprintf( fout_model_other, "num test docs: %zd\n", _D_test);
	fprintf( fout_model_other, "num test words: %zd\n", _V_test);
	fprintf( fout_model_other, "num topics: %zd\n", _K);

	/////////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////// write the reduced vocab index here
	for ( size_t i = 0; i < _test_datum_id_vec.size(); ++i )
		fprintf( fout_reducedvocab_other, "%zd=%zd\n", i, _test_datum_id_vec[i] );

	/////////////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////////////////////////////////////////////// write model assign here
	size_t n_terms, term_count;
	for ( size_t d = 0; d < _D_test; ++d ) {
		vector<int> *term_id_vec = _corpus_test->_document_vec->at(d)->_datum_id_vec;
		vector<int> *term_id_count_vec = _corpus_test->_document_vec->at(d)->_datum_id_count_vec;
		n_terms = term_id_vec->size();
		for ( size_t n = 0; n < n_terms; ++n ) {
			term_count = term_id_count_vec->at(n);
			for ( size_t c = 0; c < term_count; ++c )
				fprintf( fout_model_word_assign, " %d:%d", term_id_vec->at(n), (*(*_z_d_n_test)[d])[n] );
		}
		fprintf( fout_model_word_assign, "\n"); fflush(fout_model_word_assign);	// flushed and written a full document
	}
	// closing descriptors for custom files
	if ( fout_model_other ) fclose( fout_model_other );
	if ( fout_model_word_assign ) fclose( fout_model_word_assign );
	if ( fout_reducedvocab_other ) fclose( fout_reducedvocab_other );

	/////////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////// write parameters and hidden variables here
	CParameterMatrixReaderWriter pmrw;
	// theta
	pmrw.write_parameter_matrix_in_text_mode(model_theta_filename, _theta_test);
	// beta
	pmrw.write_parameter_matrix_in_text_mode(model_beta_filename, _beta_test);
	if ( b_verbose_global ) cout << " done." << endl;
}

/**
 *
 */
void CLDAGibbs::save_test_model_in_binary_mode() {
	// TODO to be implemented
}

/**
 * This function serves the same purpose as save_training_model_in_text_mode(string model_root_dir) except that the
 * model parameters are saved in binary files for less storage requirements
 */
void CLDAGibbs::save_training_model_in_binary_mode() {
	// TODO: to be implemented
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////// ACTUAL MODEL CREATION //////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 * what kind of data structures to create
 */
bool CLDAGibbs::create_and_initialize_data_structures() {
	if (  CLDAGeneralModel::_model_operation_mode == "train" ) {
		return create_and_initialize_data_structures_for_training();
	} else if ( CLDAGeneralModel::_model_operation_mode == "test" ) {
		return create_and_initialize_data_structures_for_testing();
	}
	return false;	// error return
}

/**
 * TRAINING
 * Creates and Initializes the data structures for the model during the training phase
 */
bool CLDAGibbs::create_and_initialize_data_structures_for_training() {
	cout << "model is creating or initializing data structures for training..." ;
	if ( _D == 0 || _V == 0 || _K == 0 ) {
		if (b_verbose_global) std::cout << "the data file " << _model_index_filename << " might be missing.. was the filepath specified correctly? OR" << endl;
		if (b_verbose_global) std::cout << "the data config file might not have been read correctly.. check config file property tags" << endl;
		return false;
	}

	CTemplateMatrixMath<int>::create_2D_matrix(&_z_d_k, _D, _K);
	CTemplateMatrixMath<int>::create_2D_matrix(&_z_v_k, _V, _K);
	_z_d_k_total = new vector<int>(_D, 0);
	_z_v_k_total = new vector<int>(_K, 0);

	CTemplateMatrixMath<double>::create_2D_matrix(&_theta, _D, _K);
	CTemplateMatrixMath<double>::create_2D_matrix(&_beta, _K, _V);

	_z_d_n = new vector<vector<int>* >(_D);
	size_t doc_id, n;
	size_t num_unique_datum, datum_id, datum_count, total_datum_count;

	/*
	 * Iterate over all documents in the dataset
	 */
	for ( doc_id = 0; doc_id < _D; ++doc_id ) {
		num_unique_datum = _corpus->_document_vec->at(doc_id)->_num_of_unique_datum;
		(*_z_d_n)[doc_id] = new vector<int>(num_unique_datum);

		total_datum_count = 0;
		// initialization for z_d_n{d}
		for ( n = 0; n < num_unique_datum; ++n ) {
			size_t topic = (size_t) ((((unsigned long) randomMT()) / 4294967296.0)*_K);	// choose a topic ~ uniform(0, K-1)
			assert(topic < _K);
			(*(*_z_d_n)[doc_id])[n] = topic;

			datum_count = _corpus->_document_vec->at(doc_id)->_datum_id_count_vec->at(n);
			datum_id = _corpus->_document_vec->at(doc_id)->_datum_id_vec->at(n);

			// number of instances of word i assigned to topic j
			(*(*_z_v_k)[datum_id])[topic] +=	datum_count;
			// total number of words assigned to topic j
			(*_z_v_k_total)[topic] += datum_count;

			// number of words in document i assigned to topic j
			(*(*_z_d_k)[doc_id])[topic] += datum_count;
			total_datum_count += datum_count;
		}

		// total number of words in document i
		(*_z_d_k_total)[doc_id] = total_datum_count;
	}

	//print_parameters();	// TODO remove
	cout << " done." << endl;
	return true;
}

/**
 * TESTING
 * Creates and Initializes the data structures for the model during the validation or testing phases
 */
bool CLDAGibbs::create_and_initialize_data_structures_for_testing() {
	cout << "model is creating or initializing data structures for validation or testing...";
	if ( _D_test == 0 || _V_test == 0 || _V == 0 || _K == 0 ) {
		if (b_verbose_global) std::cout << "\nthe data file " << _model_index_filename << " might be missing.. was the filepath specified correctly? OR" << endl;
		if (b_verbose_global) std::cout << "the data config file might not have been read correctly.. check config file property tags" << endl;
		return false;
	}
	assert( _V_test <= _V );

	CTemplateMatrixMath<int>::create_2D_matrix(&_z_d_k_test, _D_test, _K);	// vectors initialized with 0 within the function called
	CTemplateMatrixMath<int>::create_2D_matrix(&_z_v_k_test, _V_test, _K);
	_z_d_k_total_test = new vector<int>(_D_test, 0);
	_z_v_k_total_test = new vector<int>(_K, 0);

	CTemplateMatrixMath<double>::create_2D_matrix(&_theta_test, _D_test, _K);
	CTemplateMatrixMath<double>::create_2D_matrix(&_beta_test, _K, _V_test);


	_z_d_n_test = new vector<vector<int>* >(_D_test);
	size_t doc_id, n;
	size_t num_unique_test_datum, test_datum_id, test_datum_count, total_test_datum_count;

	/*
	 * Iterate over all documents in the dataset
	 */
	for ( doc_id = 0; doc_id < _D_test; ++doc_id ) {
		num_unique_test_datum = _corpus_test->_document_vec->at(doc_id)->_num_of_unique_datum;
		(*_z_d_n_test)[doc_id] = new vector<int>(num_unique_test_datum);

		total_test_datum_count = 0;
		// initialization for z_d_n{d}
		for ( n = 0; n < num_unique_test_datum; ++n ) {
			size_t topic = (size_t) ((((unsigned long) randomMT()) / 4294967296.0)*_K);;	// choose a topic ~ uniform[0, K-1]
			assert(topic < _K);
			(*(*_z_d_n_test)[doc_id])[n] = topic;

			test_datum_count = _corpus_test->_document_vec->at(doc_id)->_datum_id_count_vec->at(n);
			test_datum_id = _corpus_test->_document_vec->at(doc_id)->_datum_id_vec->at(n);

			// number of instances of word i assigned to topic j
			(*(*_z_v_k_test)[_test_datum_id_to_sequential_index_id_map[test_datum_id]])[topic] +=	test_datum_count;
			// total number of words assigned to topic j
			(*_z_v_k_total_test)[topic] += test_datum_count;

			// number of words in document i assigned to topic j
			(*(*_z_d_k_test)[doc_id])[topic] += test_datum_count;
			total_test_datum_count += test_datum_count;
		}

		// total number of words in document i
		(*_z_d_k_total_test)[doc_id] = total_test_datum_count;
	}

	//print_parameters();	// TODO remove
	cout << " done." << endl;

	/////////////////////// diagnostics /////////////////////////////
	//map<size_t, size_t>::iterator m_itr;
	//for ( m_itr = _test_datum_id_to_sequential_index_id_map.begin(); m_itr != _test_datum_id_to_sequential_index_id_map.end(); ++m_itr )
	//	if (b_verbose_global) std::cout << m_itr->first << " : " << m_itr->second << endl;
	return true;
}

/**
 * returns the training or test set likelihood depending on the model operation mode
 * @return the model likelihood for the data or numeric_limits<double>::max() if model operation mode
 * 			doesn't match "train", "training", "test", "testing", "validate", "validating" or "validation"
 */
double CLDAGibbs::compute_log_likelihood() {
	if (  CLDAGeneralModel::_model_operation_mode == "train" ) {
		return compute_log_likelihood_for_training();
	} else if ( CLDAGeneralModel::_model_operation_mode == "test" ) {
		return compute_log_likelihood_for_testing();
	}
	return numeric_limits<double>::max();	// error case
}

/**
 * This function computes the likelihood of the dataset under this model during training
 */
double CLDAGibbs::compute_log_likelihood_for_training() {
	_log_prob = 0.0;
	// add log of the priors here
	_log_prob += log(_alpha);
	_log_prob += log(_eta);

	size_t doc_id, n, topic, num_unique_datum;
	double temp_sum = 0;

	////////////////////////////////////////////
	// Iterate over all documents in the dataset
	////////////////////////////////////////////
	for ( doc_id = 0; doc_id < _D; ++doc_id ) {
		num_unique_datum = _corpus->_document_vec->at(doc_id)->_num_of_unique_datum;

		for ( n = 0; n < num_unique_datum; ++n ) {
			//topic = (*(*_z_d_n)[doc_id])[n];

			temp_sum = 0.0;
			for ( topic = 0; topic < _K; ++topic)
				temp_sum += ( (*(*_theta)[doc_id])[topic] * (*(*_beta)[topic])[_corpus->_document_vec->at(doc_id)->_datum_id_vec->at(n)] );
			_log_prob += log(temp_sum);
		}
	}
	/*
	size_t v;
	//////////////////// Griffiths and Steyver's likelihood
	_log_prob += _K*lgamma(_V*_eta);
	_log_prob -= _K*_V*lgamma(_eta);
	for ( v = 0; v < _V; ++v ) {
		for ( size_t k = 0; k < _K; ++k ) {
			_log_prob += lgamma(_z_v_k->at(v)->at(k)+_eta);
		}
	}
	for ( size_t k = 0; k < _K; ++k ) {
		_log_prob -= lgamma(_z_v_k_total->at(k) + _eta*_V);
	}*/
	///////////////////////////////////////////////////////

	return _log_prob;
}

/**
 * This function computes the likelihood of the dataset under this model during validation or testing
 */
double CLDAGibbs::compute_log_likelihood_for_testing() {
	_log_prob = 0.0;
	// add log of the priors here
	_log_prob += log(_alpha);
	_log_prob += log(_eta);

	size_t doc_id, n, topic, num_unique_datum, datum_id_test;
	double temp_sum = 0;
	size_t v_id_test;		// id of the test datum in the vocabulary

	////////////////////////////////////////////
	// Iterate over all documents in the dataset
	////////////////////////////////////////////
	for ( doc_id = 0; doc_id < _D_test; ++doc_id ) {
		num_unique_datum = _corpus_test->_document_vec->at(doc_id)->_num_of_unique_datum;

		for ( n = 0; n < num_unique_datum; ++n ) {
			//topic = (*(*_z_d_n_test)[doc_id])[n];

			temp_sum = 0.0;
			datum_id_test = _corpus_test->_document_vec->at(doc_id)->_datum_id_vec->at(n);
			v_id_test = _test_datum_id_to_sequential_index_id_map[datum_id_test];
			for ( topic = 0; topic < _K; ++topic)
				temp_sum += ( (*(*_theta_test)[doc_id])[topic] * (*(*_beta_test)[topic])[v_id_test] );
			_log_prob += log(temp_sum);
		}
	}
	/*
	size_t v;
	//////////////////// Griffiths and Steyver's likelihood
	_log_prob += _K*lgamma(_V_test*_eta);
	_log_prob -= _K*_V_test*lgamma(_eta);
	for ( v = 0; v < _V_test; ++v ) {
		for ( size_t k = 0; k < _K; ++k ) {
			_log_prob += lgamma(_z_v_k_test->at(v)->at(k)+_eta);
		}
	}
	for ( size_t k = 0; k < _K; ++k ) {
		_log_prob -= lgamma(_z_v_k_total_test->at(k) + _eta*_V_test);
	}*/
	///////////////////////////////////////////////////////

	return _log_prob;
}

/**
 * @brief compute entropy of probability density vector i.e. compute -p_i \log p_i over all i
 * @param p is the vector of doubles representing the probability density vector
 * @return the divergence value as a double
 */
double CLDAGibbs::compute_entropy(vector<double>* p) {
	double score = 0;
	size_t dist_length = p->size();
	/*
	 * entropy of p = \sum_{x \in X} -p(x) \log p(x)
	 */
	for ( size_t i = 0; i < dist_length; ++i ) {
		if ( (p->at(i) != 0) )
			score -= p->at(i)*log2(p->at(i));
	}
	return score;
}

/**
 * @brief computes the Bayesian Information Criterion from the model for the particular setting of model parameters
 * @return the Bayesian Information Criterion from the model for the particular setting of model parameters
 *
 * By default returns 0 from base class
 */
double CLDAGibbs::get_BIC() {
	return 0.5 * ((_K-1)*(_D) + (_K)*(_V-1)) * log2(_training_datum_total_count);
}
/**
 * @brief computes the Minimum Description Length from the model for the particular setting of model parameters
 * @return the Minimum Description Length from the model for the particular setting of model parameters
 *
 * By default returns 0 from base class
 */
double CLDAGibbs::get_MDL() {
	double entropy = 0;

	for ( size_t d = 0; d < _D; ++d )
		entropy += compute_entropy(_theta->at(d));
	for ( size_t k = 0; k < _K; ++k )
		entropy += compute_entropy(_beta->at(k));

	/*
	entropy = 0.5 * ((_K-1)*(_D) + (_K)*(_V-1)) * log2
			(
					((_K-1)*(_D) + (_K)*(_V-1)) * sizeof(double)*8
			);
	*/
	return entropy;
}
