/*
 * CModelGraph.cpp
 *
 *  Created on: May 19, 2010
 *      Author: pdas
 */

#include "CModelGraph.h"

using namespace std;

void CModelGraph::add_to_conjugate_distribution_multimap() {
	//////////////////////////////////////////////////////////
	//  adding a few conjugate distribution identifiers     //
	_conjugate_distribution_multimap.insert( make_pair("dirsym","mult") );
	_conjugate_distribution_multimap.insert( make_pair("dir","mult") );
	_conjugate_distribution_multimap.insert( make_pair("nor","nor") );
	_conjugate_distribution_multimap.insert( make_pair("gam","exp") );
	_conjugate_distribution_multimap.insert( make_pair("beta","bino") );
	_conjugate_distribution_multimap.insert( make_pair("gam","dirsym") );
	_conjugate_distribution_multimap.insert( make_pair("gam","poison") );
	_conjugate_distribution_multimap.insert( make_pair("ig","nor") );
}

CModelGraph::CModelGraph() {

	_model_bytecode = "";

	////////////////////////////////////////////////////////////////////////////////////////////
	//					The following three statements are just instantiations				  //
	_model_graph = new map<SModelGraphNode*, list<SModelGraphNode*>*, CModelGraphNodeCompare >();
	_observed_variable_set = new set<SModelGraphNode*, CModelGraphNodeCompare>();
	_model_forest_representation = new SForest();

	add_to_conjugate_distribution_multimap();

}

/**
 * copy constructor
 */
CModelGraph::CModelGraph(CModelGraph& src_model_graph) {

	_model_bytecode = "";

	////////////////////////////////////////////////////////////////////////////////////////////
	//					The following three statements are just instantiations				  //
	_model_graph = new map<SModelGraphNode*, list<SModelGraphNode*>*, CModelGraphNodeCompare >();
	_observed_variable_set = new set<SModelGraphNode*, CModelGraphNodeCompare>();
	_model_forest_representation = new SForest();

	//////////////////////////////////////////////////////////
	//  adding a few conjugate distribution identifiers     //
	add_to_conjugate_distribution_multimap();

	deeply_copy_graphical_model(&src_model_graph);

}

/**
 * assignment operator
 */
CModelGraph& CModelGraph::operator = (CModelGraph& src_model_graph ) {
	_model_bytecode = "";

	////////////////////////////////////////////////////////////////////////////////////////////
	//					The following three statements are just instantiations				  //
	_model_graph = new map<SModelGraphNode*, list<SModelGraphNode*>*, CModelGraphNodeCompare >();
	_observed_variable_set = new set<SModelGraphNode*, CModelGraphNodeCompare>();
	_model_forest_representation = new SForest();

	//////////////////////////////////////////////////////////
	//  adding a few conjugate distribution identifiers     //
	add_to_conjugate_distribution_multimap();

	deeply_copy_graphical_model(&src_model_graph);

	return *this;
}

/**
 * Model Graph destructor
 */
CModelGraph::~CModelGraph() {
	map<SModelGraphNode*, list<SModelGraphNode*>*, CModelGraphNodeCompare>::iterator graph_itr;
	for ( graph_itr = _model_graph->begin(); graph_itr != _model_graph->end(); ++graph_itr ) {
		delete graph_itr->second;
	}
	delete _model_graph;
	delete _model_forest_representation;
	delete _observed_variable_set;
}

/**
 * This method prints out a string representation of the concrete underlying graph
 */
string CModelGraph::to_string() {
	string s = "[ \n";
	map<SModelGraphNode*, list<SModelGraphNode*>*, CModelGraphNodeCompare>::iterator graph_itr;
	list<SModelGraphNode*>::iterator list_itr;
	for ( graph_itr = _model_graph->begin(); graph_itr != _model_graph->end(); ++graph_itr ) {
		s += graph_itr->first->to_string() + " | ";
		list<SModelGraphNode*> *adjacency_list = graph_itr->second;
		for ( list_itr = adjacency_list->begin(); list_itr != adjacency_list->end(); ++list_itr )
			s += (*list_itr)->to_string() + " -> ";
		s += '\n';
	}
	s += " ]";
	return s;
}

/**
 * This function just prints the to_string representation of the nodes in the graph
 */
void CModelGraph::print_graph() const {
	map<SModelGraphNode*, list<SModelGraphNode*>*, CModelGraphNodeCompare>::iterator graph_itr;
	list<SModelGraphNode*>::iterator list_itr;
	for ( graph_itr = _model_graph->begin(); graph_itr != _model_graph->end(); ++graph_itr ) {
		if (b_verbose_global) std::cout << graph_itr->first->to_string() << " | ";
		list<SModelGraphNode*> *adjacency_list = graph_itr->second;
		for ( list_itr = adjacency_list->begin(); list_itr != adjacency_list->end(); ++list_itr )
			if (b_verbose_global) std::cout << (*list_itr)->to_string() << "->";
		if (b_verbose_global) std::cout << endl;
	}
}

/**
 * This function returns a deep copy of the source graph model and stores the pointer to the newly
 * created graph in the instance variable _model_graph
 */
void CModelGraph::deeply_copy_graphical_model(CModelGraph* src_graphical_model) {

	map<SModelGraphNode*, list<SModelGraphNode*>*, CModelGraphNodeCompare>* src_model_graph = src_graphical_model->get_graph();

	map<SModelGraphNode*, list<SModelGraphNode*>*, CModelGraphNodeCompare>::iterator src_graph_itr;
	map<SModelGraphNode*, list<SModelGraphNode*>*, CModelGraphNodeCompare>::iterator this_graph_itr, graph_itr;
	list<SModelGraphNode*>::iterator list_itr;

	SModelGraphNode temp_node;

	for ( src_graph_itr = src_model_graph->begin(); src_graph_itr != src_model_graph->end(); ++src_graph_itr ) {

		SModelGraphNode* new_node = new SModelGraphNode(*(src_graph_itr->first));	// invoking default copy constructor on Node
		list<SModelGraphNode*> *new_list = new list<SModelGraphNode*>();

		// maintain the observed variable set
		if ( new_node->is_observed() ) _observed_variable_set->insert(new_node);

		this_graph_itr = _model_graph->find(new_node);
		if ( this_graph_itr == _model_graph->end() )
			_model_graph->insert( make_pair(new_node, new_list) );

		//if (b_verbose_global) std::cout << new_node->_variable_name << " .. ";
		list<SModelGraphNode*> *adjacency_list = src_graph_itr->second;
		for ( list_itr = adjacency_list->begin(); list_itr != adjacency_list->end(); ++list_itr ) {
			temp_node = *(*list_itr);	// repeatedly copy into one node for searching -- search with "&temp_node"

			//if (b_verbose_global) std::cout << temp_node._variable_name ;
			this_graph_itr = _model_graph->find(&temp_node);
			if ( this_graph_itr == _model_graph->end() ) {
				//if (b_verbose_global) std::cout << " not found.." ;
				SModelGraphNode* future_node = new SModelGraphNode(temp_node);
				// insert a new node in the new graph
				_model_graph->insert( make_pair(future_node, new list<SModelGraphNode*>()) );
				////////////////////////////////////////////////////////////////////////////////
				// also add the pointer to the newly inserted node to the current node's list //
				graph_itr = _model_graph->find(new_node);
				if ( graph_itr != _model_graph->end() )
					graph_itr->second->push_back(future_node);
				else if (b_verbose_global) std::cout << "FATAL error!!! graph_itr = new_graph->find(new_node); couldn't find an existing node" << endl;

			} else {
				//if (b_verbose_global) std::cout << " found.. " << new_graph_itr->first->_variable_name << "(exists)  " ;
				///////////////////////////////////////////////////////////////////////////
				// add the pointer to the newly inserted node to the current node's list //
				graph_itr = _model_graph->find(new_node);
				if ( graph_itr != _model_graph->end() )
					graph_itr->second->push_back(this_graph_itr->first);
				else if (b_verbose_global) std::cout << "FATAL error!!! graph_itr = new_graph->find(new_node); couldn't find an existing node" << endl;

			}

		}

		//if (b_verbose_global) std::cout << endl;
	}

	// builds the model forest representation
	this->level_order_bfs(_model_forest_representation);
	return ;
}

/**
 *
 */
void CModelGraph::build_complete_graph() {
	string variable_name;
	SModelGraphNode* node;
	map<SModelGraphNode*, list<SModelGraphNode*>*, CModelGraphNodeCompare>::iterator graph_itr, graph_search_itr;
	list<string>::iterator list_itr;

	SModelGraphNode temp_node;

	for ( graph_itr = _model_graph->begin(); graph_itr != _model_graph->end(); ++graph_itr ) {
		node = graph_itr->first;
		variable_name = node->_variable_name;
		list<SModelGraphNode*>* adjacency_list = graph_itr->second;

		// maintain the observed variable set
		if ( node->is_observed() ) _observed_variable_set->insert(node);

		list<string>& gen_dist_string_param_list = node->_generating_distribution_string_parameters_list;
		for ( list_itr = gen_dist_string_param_list.begin(); list_itr != gen_dist_string_param_list.end(); ++list_itr ) {
			temp_node.set_variable_name(*list_itr);

			graph_search_itr = _model_graph->find(&temp_node);
			if ( graph_search_itr != _model_graph->end() )
				adjacency_list->push_back(graph_search_itr->first);
		}
	}

	this->level_order_bfs(_model_forest_representation);	// create the forest from the graphical model graph
}

/**
 * This function reverses the graph
 * Actually we don't need to do this since the adjacency list is already in reverse order :-)
 * in the current graph, we have w | beta -> z
 */
void CModelGraph::reverse_graph() {
	// do the stuff with _model_graph

	// can a graph with adjacency list representation be reversed in place? No, right?
}

/**
 * This function does a level order breadth first traversal for each observed variable
 * returns a new forest of trees
 *
 * The forest is structured as map<string, map<int, set<string>*>*>
 * 									^			^			^
 * 						observed variable_name/ level child distribution_name set
 */
void CModelGraph::level_order_bfs(SForest* model_forest_representation) {

	set<SModelGraphNode*, CModelGraphNodeCompare>::iterator set_itr;

	// original graph iterator
	map<SModelGraphNode*, list<SModelGraphNode*>*, CModelGraphNodeCompare>::iterator graph_itr, graph_search_itr;

	if (b_verbose_global) std::cout << "observed variable set size = " << _observed_variable_set->size() << endl;
	// building the forest
	for ( set_itr = _observed_variable_set->begin(); set_itr != _observed_variable_set->end(); ++set_itr ) {
		string observed_variable_name = (*set_itr)->_variable_name;

		map<int, set<string>* >* level_order_tree = 0;

		graph_search_itr = _model_graph->find(*set_itr);
		if ( graph_search_itr != _model_graph->end() ) {
			level_order_tree = level_order_bfs_subgraph(_model_graph, graph_search_itr->first);

		} else {
			if (b_verbose_global) std::cout << "FATAL error: observed variable not found in model graph!" << endl;
		}

		// finally insert into the forest
		model_forest_representation->_forest->insert( make_pair(observed_variable_name+"(observed)", level_order_tree) );
	}

	return;
}


/**
 * populate one tree under the observed variable in the forest using level order bfs
 */
map<int, set<string>* >* CModelGraph::level_order_bfs_subgraph(
									map<SModelGraphNode*, list<SModelGraphNode*>*, CModelGraphNodeCompare>* model_graph,
									SModelGraphNode* start) {

	map<int, set<string>* >* level_order_tree = new map<int, set<string>* >();

	const int BLACK = 0;
	const int GRAY = 1;
	const int WHITE = 2;

	map<SModelGraphNode*, list<SModelGraphNode*>*, CModelGraphNodeCompare>::iterator graph_search_itr;

	queue<SModelGraphNode*> q;
	q.push(start);

	// info local to working of bfs graph traversal only.
	SNodeInfo ni;
	map<SModelGraphNode*, SNodeInfo> m;	// predecessor tracking map
	map<SModelGraphNode*, SNodeInfo>::iterator m_itr;

	ni._depth = 0;
	ni._predecessor = 0;
	ni._visit_status = WHITE;
	m[start] = ni;

	// loop while queue not empty.
	while (!q.empty()) {
		SModelGraphNode *predecessor = q.front();
		m_itr = m.find(predecessor);
		if ((m_itr != m.end()) && m_itr->second._visit_status == WHITE)
			m_itr->second._visit_status = GRAY;
		q.pop();

		graph_search_itr = model_graph->find(predecessor);
		list<SModelGraphNode*>::iterator itr = graph_search_itr->second->begin();
		// the list may be empty if the variable is generated by the highest level hyperparameter
		if (graph_search_itr->second->size() != 0) {
			while (itr != graph_search_itr->second->end()) {
				q.push(*itr);
				ni._predecessor = predecessor;
				ni._depth = m_itr->second._depth + 1;
				ni._visit_status = WHITE;
				m[*itr] = ni; // insert current neighbor into the tracking map.
				++itr;
			}
			m_itr->second._visit_status = BLACK;
		}
	}

	/////////////////////////////////////
	// populating the level order tree //
	map<int, set<string>* >::iterator tree_itr;
	string generating_distribution_string;

	/*
	 * at this point we have the predecessor map m for all nodes visited under the observed variable node
	 * in a level order fashion
	 * so, we iterate through that predecessor tracking map and append generating distribution variables at each level
	 */
	for ( m_itr = m.begin(); m_itr != m.end(); ++m_itr ) {
		SModelGraphNode* node = m_itr->first;
		int level = m_itr->second._depth;
		tree_itr = level_order_tree->find(level);
		generating_distribution_string = node->_generating_distribution;

		/*
		 * the generating distribution string should look like
		 * dirsym(variable_name)(total_dimensionality)(parameter)
		 * e.g. dirsym(beta)(kv)(0.01) or dirsym(theta)(dk)(0.01)
		 */
		generating_distribution_string += "("+node->_variable_name+")";
		if (node->_observed) generating_distribution_string += "(observed)";
		else generating_distribution_string += "(hidden)";
		generating_distribution_string += "("+node->_variable_indices+")";

		list<double>& numeric_param_list = node->_generating_distribution_numeric_parameters_list;
		if ( numeric_param_list.size() != 0 ) {
			////// -- remove
			//list<double>::iterator litr;
			//if (b_verbose_global) std::cout << generating_distribution_string ;
			//for (litr = numeric_param_list.begin(); litr != numeric_param_list.end(); ++litr )
			//	if (b_verbose_global) std::cout << "* " << *litr;
			//if (b_verbose_global) std::cout << endl;
			////// -- remove
			add_numeric_parameters_to_generating_random_variable(numeric_param_list, generating_distribution_string);

		}

		list<string>& nonnumeric_param_list = node->_generating_distribution_string_parameters_list;
		if ( nonnumeric_param_list.size() != 0 ) {
			add_nonnumeric_parameters_to_generating_random_variable(node->_generating_distribution_string_parameters_list, generating_distribution_string);
		}

		if ( tree_itr == level_order_tree->end() ) {
			set<string>* s = new set<string>();
			s->insert(generating_distribution_string);
			level_order_tree->insert( make_pair(level,s) );
		} else {
			//add_numeric_parameters_to_generating_random_variable(numeric_param_list, generating_distribution_string);
			tree_itr->second->insert(generating_distribution_string);
		}
	}

	/////////////////////////////////////
	// printing the level order tree   //
//	set<string>::iterator set_itr;
//	if (b_verbose_global) std::cout << "printing the level order tree" << endl;
//	for ( tree_itr = level_order_tree->begin(); tree_itr != level_order_tree->end(); tree_itr++ ) {
//		if (b_verbose_global) std::cout << "level " << tree_itr->first << ": " ;
//		for ( set_itr = tree_itr->second->begin(); set_itr != tree_itr->second->end(); set_itr++ )
//			if (b_verbose_global) std::cout << *set_itr << ", " ;
//		if (b_verbose_global) std::cout << endl;
//	}

	return level_order_tree;
}

/**
 * auxilliary function to add numbers to a list thus appending the string with something like "(xx.yy,xxx.yyy,...)"
 */
void CModelGraph::add_numeric_parameters_to_generating_random_variable(list<double>& l, string& s) {
	list<double>::iterator l_itr;
	size_t l_size = l.size();
	size_t i = 0;
	s += "(";
	for ( l_itr = l.begin(); l_itr != l.end(); ++l_itr ) {
		s += CUtilities::number_to_string(*l_itr);
		if ( i < l_size - 1 )
			s += ";";
		++i;
	}
	s += ")";
}

/**
 * auxilliary function to add numbers to a list thus appending the string with something like "(xx.yy,xxx.yyy,...)"
 */
void CModelGraph::add_nonnumeric_parameters_to_generating_random_variable(list<string>& l, string& s) {
	list<string>::iterator l_itr;
	size_t l_size = l.size();
	size_t i = 0;
	s += "(";
	for ( l_itr = l.begin(); l_itr != l.end(); ++l_itr ) {
		s += *l_itr;
		if ( i < l_size - 1 )
			s += ";";
		++i;
	}
	s += ")";
}

/**
 * This function returns a string identifying a model for e.g. "lda", "author-topic", "entity-lda", "hmm-lda" or
 * the empty string "" if no match is obtained
 * This feature differentiates this project from HBC where the user gets feedback on what kind of model was supplied
 * Also, efficient model specific (for e.g., threaded or parallelized) code can be written without the need for complex
 * HB compiler generated code optimization
 *
 * The model forest representation is in the instance variable _model_forest_representation
 */
string CModelGraph::match_compiled_template(string compiled_model_filename) {

	string line;
	map<string, string> compiled_model_map;		// map<model_bytecode, model_name>
	map<string, string>::iterator model_map_itr;

	/////////////// read compiled model template file
	vector<string> tokens;
	ifstream fs(compiled_model_filename.c_str());
	if ( fs.is_open() ) {

		while ( !fs.eof() ) {
			getline(fs, line);

			if ( !line.length() ) continue;

			tokens.clear();

			CUtilities::tokenize(line, tokens, "=");
			CUtilities::trim_left_right(tokens[0]);
			CUtilities::to_lower(tokens[0]);
			CUtilities::trim_left_right(tokens[1]);

			compiled_model_map.insert( make_pair(tokens[1], tokens[0]) );
		}
		fs.close();
	} else { if (b_verbose_global) std::cout << "FATAL error: Couldn't open " << compiled_model_filename << endl; exit(APPLICATIONEXIT); }
	////////////// end reading compiled model template file

	string matched_model_name = "";
	map<string, map<int, set<string>*>*>::iterator f_itr;
	map<int, set<string>*>::iterator t_itr, t_next_itr;

	int level;

	string s;	// this is the compiled string representation of the model
	set<string, SStringSetCompareRemoveParenthesis> prev_level_set, level_set, current_level_conjugacy_set;
	set<string, SStringSetCompareRemoveParenthesis>::iterator level_s_itr, prev_level_s_itr, cls_itr;
	set<string>::iterator prev_s_itr, s_itr;

	// used for searching through the conjugate distributions map
	multimap<string, string>::iterator multimap_itr;
	pair<multimap<string, string>::iterator,multimap<string, string>::iterator> ret;

	/*
	 * The observed variable forest - useful for comparing different graphical models to some compiled ones
	 * essetially a map<string,map<int, set<string>*>*>* with a structure breakdown like :
	 * [map<observed_variable_name, m>]
	 * w(observed)>
	 * [m = map<    int, {string1,.., stringN} >]
	 *		[level] 0 : mult,
	 *		[level] 1 : dirsym(xx.yy,..), mult,
	 *		[level] 2 : dirsym(xx.yy,..),
	 *
	 *	Each observed variable should form the root of a tree in the _forest
	 *	Note: before checking, we need to remove the (xx.yy,..)'s
	 *
	 */
	string common_gen_dist_str;
	for (f_itr = _model_forest_representation->_forest->begin(); f_itr != _model_forest_representation->_forest->end(); ++f_itr) {

		map<int, set<string>*>* tree = f_itr->second;
		t_itr = tree->begin();
		if ( t_itr != tree->end() ) {
			t_next_itr = tree->begin(); ++t_next_itr;	// next level tree iterator
		}

		level = t_itr->first;

		if ( t_itr->second->size() > 1 ) { if (b_verbose_global) std::cout << "Warning! This observed variable is being generated by two distributions.. only the first is used" << endl;}

		// capturing the root level

		common_gen_dist_str = delete_paranthesized_expression_from_bytecode_string(*(t_itr->second->begin()));
		//s += "level" + CUtilities::number_to_string<int>(level) + ":" + common_gen_dist_str + '$';
		s += common_gen_dist_str + '$';

		while (t_next_itr != tree->end()) {

			current_level_conjugacy_set.clear();
			prev_level_set.clear();
			level_set.clear();

			level = t_next_itr->first;
			//s += "level" + CUtilities::number_to_string<int>(level) + ":";

			////////////////////////////////////////////////////////////////////////////////////////
			// need to have the conjugacy structure here
			// first copy the current and the preceding sets of strings
			for ( prev_s_itr = t_itr->second->begin(); prev_s_itr != t_itr->second->end(); ++prev_s_itr ) {
				//common_gen_dist_str = delete_paranthesized_expression_from_bytecode_string(*prev_s_itr);
				//prev_level_set.insert(common_gen_dist_str);
				prev_level_set.insert(*prev_s_itr);
			}
			for ( s_itr = t_next_itr->second->begin(); s_itr != t_next_itr->second->end(); ++s_itr ) {
				//common_gen_dist_str = delete_paranthesized_expression_from_bytecode_string(*s_itr);
				//level_set.insert(common_gen_dist_str);
				level_set.insert(*s_itr);
			}
			// note at this point *prev_s_itr and *s_itr still contains the original unmodified ()-ized expressions
			////////////////////////////////////////////////////////////////////////////////////////

			//if (b_verbose_global) std::cout << "current level = " << level << endl;
			//if (b_verbose_global) std::cout << "prev_set size = " << prev_level_set.size() << endl;
			//if (b_verbose_global) std::cout << "current_set size = " << level_set.size() << endl;

			for ( level_s_itr = level_set.begin(); level_s_itr != level_set.end(); ++level_s_itr ) {
				common_gen_dist_str = delete_paranthesized_expression_from_bytecode_string(*level_s_itr);
				ret = _conjugate_distribution_multimap.equal_range(common_gen_dist_str);

				for ( multimap_itr = ret.first; multimap_itr != ret.second; ++multimap_itr ) {
					if (b_verbose_global) std::cout << *level_s_itr << " -- " << (*multimap_itr).first << "|" << (*multimap_itr).second << endl;
					prev_level_s_itr = prev_level_set.find((*multimap_itr).second);
					if ( prev_level_s_itr != prev_level_set.end() ) {
						current_level_conjugacy_set.insert(*level_s_itr+"|"+*prev_level_s_itr);
						break;
					}
				}

			}
			if (b_verbose_global) std::cout << "conj_set size = " << current_level_conjugacy_set.size() << endl;
			size_t i = 1;
			for ( cls_itr = current_level_conjugacy_set.begin(); cls_itr != current_level_conjugacy_set.end(); ++cls_itr ) {
				s += *cls_itr;
				if ( i < current_level_conjugacy_set.size() ) s +=",";
				i++;
			}
			s += '$';
			if (b_verbose_global) std::cout << "---------" << endl;

			 ++t_itr;
			 ++t_next_itr;
		}

	}

	if (b_verbose_global) std::cout << "s = " << s << endl;
	// at this point we can remove all parenthesized expression from s like "(xx.yy,xxx.yyy)"
	string new_s = delete_paranthesized_expression_from_bytecode_string(s);

	model_map_itr = compiled_model_map.find(delete_paranthesized_expression_from_bytecode_string(s));
	_model_bytecode = s;
	if ( model_map_itr != compiled_model_map.end() ) {
		matched_model_name = model_map_itr->second;
	} else {
		matched_model_name = "unrecognized";
	}

	//if (b_verbose_global) std::cout << "model = " << s << endl;
	//if (b_verbose_global) std::cout << "matched model name = " << matched_model_name << endl;

	return matched_model_name;
}

string CModelGraph::delete_paranthesized_expression_from_bytecode_string(const string& s) {
	string new_s = "";
	size_t length = s.length();
	bool paren_found = false;
	for ( size_t i = 0; i < length; ++i ) {
		if ( s.at(i) == '(' ) {
			paren_found = true;
		}
		if ( s.at(i) == ')' && paren_found ) {
			paren_found = false;
			continue;
		}
		if ( !paren_found )
			new_s += s.at(i);
	}
	return new_s;

}

/**
 * @brief This function generates a vector of all possible paths from observed variables to priors/hyperpriors
 *
 * @param model_bytecode as a string
 * @see CLDAGeneralModel::_model_bytecode
 * @return list<list<list<SModelGraphSimpleNode> > > - Each observed variable can have a list of paths and each path is a list of nodes
 */
list< pair<SModelGraphSimpleNode, list<list<SModelGraphSimpleNode> > > > CModelGraph::get_all_paths_from_all_obs_vars(string model_bytecode) {

	vector<string> level_vec, dists_vec, dist_pair_vec, dist_attr_vec;
	string level_delim = "$", dist_delim = ",", dist_pair_delim = "|", dist_attr_delim = "()";

	/////////////////////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////////////////////////// data structures for building the simple model graph
	/////////////////////////////////////////////////////////////////////////////////////////////////////
	map<SModelGraphSimpleNode, list<SModelGraphSimpleNode> > model_simple_graph;
	map<SModelGraphSimpleNode, list<SModelGraphSimpleNode> >::iterator smpl_grph_itr, smpl_grph_lkup_itr;
	list<SModelGraphSimpleNode>::iterator smpl_l_itr;
	set<SModelGraphSimpleNode> observed_variable_set;
	set<SModelGraphSimpleNode>::iterator ovs_itr;

	////// full model graph bytecode
	CUtilities::tokenize(model_bytecode, level_vec, level_delim);
	size_t num_levels = level_vec.size();
	for ( size_t l = 0; l < num_levels; ++l ) {
		//if (b_verbose_global) std::cout << level_vec[l] << endl;

		////// pairs of conjugate distribution
		dists_vec.clear();
		CUtilities::tokenize(level_vec[l], dists_vec, dist_delim);
		size_t num_dists = dists_vec.size();
		for ( size_t i = 0; i < num_dists; ++i ) {
			//if (b_verbose_global) std::cout << " " << dists_vec[i] << endl;

			////// individual distributions
			dist_pair_vec.clear();
			CUtilities::tokenize(dists_vec[i], dist_pair_vec, dist_pair_delim);
			if ( dist_pair_vec.size() != 2 ) continue;

			// for this pair of r.v's make a edge from the second to the first (reverse of the generative direction)
			size_t num_pairs = dist_pair_vec.size();
			SModelGraphSimpleNode gen_dist_node;

			for ( size_t j = 0; j < num_pairs; ++j ) {
				//if (b_verbose_global) std::cout << "   " << dist_pair_vec[j] << "{";

				////// each distribution attribute vector
				dist_attr_vec.clear();
				CUtilities::tokenize(dist_pair_vec[j], dist_attr_vec, dist_attr_delim);
				size_t num_attrs = dist_attr_vec.size();

				/*
				 * Each node must have the generating distribution name, it's name, r.v. type (hidden or observed) and dimensionality
				 */
				if ( num_attrs > 3 ) {

					SModelGraphSimpleNode a_node(dist_attr_vec[1]);
					if ( dist_attr_vec[2] == "observed" )
						a_node._observed = true;
					a_node._dimensionality = dist_attr_vec[3];

					list<SModelGraphSimpleNode> node_adj_list;
					//for ( size_t k = 0; k < num_attrs; ++k ) {
					//	if (b_verbose_global) std::cout << "    " << dist_attr_vec[k] ;
					//}
					//if (b_verbose_global) std::cout << "}" << endl;
					for ( size_t k = 4; k < num_attrs; ++k )
						a_node._parameter_list.push_back(dist_attr_vec[k]);
					if ( a_node._observed ) observed_variable_set.insert( a_node );

					if ( j == 0 ) gen_dist_node = a_node;	// copy a_node for lookup
					if ( j == 1 ) node_adj_list.push_back(gen_dist_node);

					// insert into the model_simple_graph
					smpl_grph_itr = model_simple_graph.find(a_node);
					if ( smpl_grph_itr == model_simple_graph.end() ) {
						model_simple_graph.insert( make_pair(a_node, node_adj_list) );
					}
				} else {
					if (b_verbose_global) std::cout << "cannot decode model bytecode for " << dist_attr_vec[j] << endl;
					continue;
				}

			}	// each pair is processed here
		}
	}

	//////////////////// DIAGNOSTICS
	/////// print model_simple_graph
	/*SModelGraphSimpleNode n;
	for ( smpl_grph_itr = model_simple_graph.begin(); smpl_grph_itr != model_simple_graph.end(); ++smpl_grph_itr ) {
		n = smpl_grph_itr->first;
		if (b_verbose_global) std::cout <<  n.to_string() << " [ ";
		for ( smpl_l_itr = smpl_grph_itr->second.begin(); smpl_l_itr != smpl_grph_itr->second.end(); ++smpl_l_itr )
			if (b_verbose_global) std::cout << " " << (*smpl_l_itr).to_string();
		if (b_verbose_global) std::cout << " ] " << endl;
	}*/

	list< pair<SModelGraphSimpleNode, list<list<SModelGraphSimpleNode> > > > all_paths;
	/////// iterate over all the observed variables in the set
	for ( ovs_itr = observed_variable_set.begin(); ovs_itr != observed_variable_set.end(); ++ovs_itr ) {
		// find the observed variable in the graph
		smpl_grph_itr = model_simple_graph.find(*ovs_itr);
		if ( smpl_grph_itr == model_simple_graph.end() ) {
			if (b_verbose_global) std::cout << "Very strange!!! observed variable with name " << (*ovs_itr)._variable_name << " cannot be found in the graph!!!" << endl; continue;
		}
		list<list<SModelGraphSimpleNode> > paths;
		SModelGraphSimpleNode node = *ovs_itr;
		//if (b_verbose_global) std::cout << "DEBUG: generating all paths.. " << endl;
		find_paths(model_simple_graph, node, paths);
		all_paths.push_back( make_pair(node, paths) );
	}

	return all_paths;
}

/**
 *
 */
void CModelGraph::find_paths(
		map<SModelGraphSimpleNode, list<SModelGraphSimpleNode> >& model_simple_graph,
		SModelGraphSimpleNode& node,
		list<list<SModelGraphSimpleNode> >& paths) {

	map<SModelGraphSimpleNode, list<SModelGraphSimpleNode> >::iterator smpl_grph_itr, smpl_grph_lkup_itr;
	list<SModelGraphSimpleNode>::iterator smpl_l_itr;

	// grab the observed variable node - there is always going to be one
	smpl_grph_lkup_itr = model_simple_graph.find(node);
	SModelGraphSimpleNode& obs_node = const_cast<SModelGraphSimpleNode& >(smpl_grph_lkup_itr->first);
	list<SModelGraphSimpleNode>& adj_list = smpl_grph_lkup_itr->second;

	if ( !adj_list.size() )	// an observed variable must be generated!!!
		return;

	vector<string> param_decomp_vec;
	string param_decomp_delim = ";";
	list<string>::iterator l_itr;

	string dependency;
	list<string>& param_list = obs_node._parameter_list;
	for ( l_itr = param_list.begin(); l_itr != param_list.end(); ++l_itr ) {
		param_decomp_vec.clear();
		CUtilities::tokenize(*l_itr, param_decomp_vec, param_decomp_delim);
		size_t num_dependencies = param_decomp_vec.size();

		for ( size_t i = 0; i < num_dependencies; i++ ) {
			dependency = param_decomp_vec[i];
			SModelGraphSimpleNode temp_node(dependency);
			//if (b_verbose_global) std::cout << dependency << " -> In find_paths " << temp_node.to_string() << endl;

			list<SModelGraphSimpleNode> a_path;
			// recursive call build_path_recursively(model_simple_graph, temp_node, a_path);
			build_path_recursively(model_simple_graph, temp_node, a_path);
			if ( !a_path.size() ); else paths.push_back(a_path);
		}
	}

	//////////////////////// DIAGNOSTICS
	/// print list<list<SModelGraphSimpleNode> >& paths
	/*if (b_verbose_global) std::cout << "finding paths of obs var = " << node._variable_name << endl;
	list<list<SModelGraphSimpleNode> >::iterator itr1;
	list<SModelGraphSimpleNode>::iterator itr2;
	if (b_verbose_global) std::cout << "[ ";
	for ( itr1 = paths.begin(); itr1 != paths.end(); ++itr1 ) {
		if (b_verbose_global) std::cout << " { " ;
		for ( itr2 = (*itr1).begin(); itr2 != (*itr1).end(); ++itr2 )
			if (b_verbose_global) std::cout << "  " << (*itr2).to_string() ;
		if (b_verbose_global) std::cout << " } " ;
	}
	if (b_verbose_global) std::cout << "]" << endl;
	*/
	////////////////////////////////////
}

void CModelGraph::build_path_recursively(
		map<SModelGraphSimpleNode, list<SModelGraphSimpleNode> >& model_simple_graph,
		SModelGraphSimpleNode& node,
		list<SModelGraphSimpleNode>& path) {

	map<SModelGraphSimpleNode, list<SModelGraphSimpleNode> >::iterator smpl_grph_itr, smpl_grph_lkup_itr;
	list<SModelGraphSimpleNode>::iterator smpl_l_itr;

	// lookup the node in the model_simple_graph
	smpl_grph_lkup_itr = model_simple_graph.find(node);
	if ( smpl_grph_lkup_itr == model_simple_graph.end() ) {
		return;
	}

	SModelGraphSimpleNode& a_node = const_cast<SModelGraphSimpleNode& >(smpl_grph_lkup_itr->first);
	list<SModelGraphSimpleNode>& adj_list = smpl_grph_lkup_itr->second;

	SModelGraphSimpleNode a_terminal_node = a_node;
	path.push_back(a_terminal_node);
	if ( !adj_list.size() ) {	// no need to explore further
		return;
	}

	vector<string> param_decomp_vec;
	string param_decomp_delim = ";";
	list<string>::iterator l_itr;

	string dependency;
	list<string>& param_list = a_node._parameter_list;
	for ( l_itr = param_list.begin(); l_itr != param_list.end(); ++l_itr ) {
		param_decomp_vec.clear();
		CUtilities::tokenize(*l_itr, param_decomp_vec, param_decomp_delim);
		size_t num_dependencies = param_decomp_vec.size();

		for ( size_t i = 0; i < num_dependencies; i++ ) {
			dependency = param_decomp_vec[i];
			SModelGraphSimpleNode temp_node(dependency);
			//if (b_verbose_global) std::cout << dependency << " -> In build_path_recursively " << temp_node.to_string() << endl;

			// recursive call build_path_recursively(model_simple_graph, temp_node, a_path);
			build_path_recursively(model_simple_graph, temp_node, path);
		}
	}

}

