/*
 * CModelGraphBuilder.cpp
 *
 *  Created on: May 17, 2010
 *      Author: pdas
 */

#include "CModelGraphBuilder.h"

using namespace std;

CModelGraphBuilder::CModelGraphBuilder() {

	//_variables_seen_so_far_set = new set<string>();
}

CModelGraphBuilder::~CModelGraphBuilder() {
	if (b_verbose_global) std::cout << "deleting graphical model...";
	if (_graphical_model) delete _graphical_model;
	//delete _conjugate_distribution_identifier_set;
	//delete _variables_seen_so_far_set;
	if (b_verbose_global) std::cout << " done" << endl;
}

void CModelGraphBuilder::instantiate_graphical_model() {
	_graphical_model = new CModelGraph();
	_model_graph = _graphical_model->get_graph();
}

/**
 * This function reads in the parsed-token string to extract the variable
 * Note: this just creates a
 */
void CModelGraphBuilder::add_node_to_graphical_model(string variable, string generating_distribution) {

	if ( !_model_graph ) { if (b_verbose_global) std::cout << "no model graph created..."; return; }

	//if (b_verbose_global) std::cout << "beginning of a node ------- " << endl;

	string observed_flag_delimiter = "@";		// the delimiters used for tokenization of @observed flag
	string zorder_index_delimiter = "_{,}";		// the delimiters used for tokenization of strings like w_{d,n}
	string parameter_index_delimiter = "()";	// the delimiters used for tokenization of strings like DirSym(...)
	vector<string> observed_tokens_vec;			// this vector is used in tokenization for obtaining observed flag for variables
	vector<string> zorder_tokens_vec;			// this vector is used in tokenization for indices
	vector<string> parameter_tokens_vec;		// this vector is used in tokenization for distribution parameters

	//////////////////////////////////////////////////
	// make all strings lower-case in the beginning //
	CUtilities::to_lower(variable);
	CUtilities::to_lower(generating_distribution);
	//////////////////////////////////////////////////

	// tokenizing the variable part for zorders (indices within the corpus)
	CUtilities::tokenize(variable, zorder_tokens_vec, zorder_index_delimiter);
	variable = zorder_tokens_vec[0];			// there should be at least one for variables like beta_{k} or alpha

	string variable_indices = "";	// indices is empty typically for hyperparameters
	for ( size_t i = 1 ; i < zorder_tokens_vec.size(); i++ ) {// this loop could be skipped if there are no parameters
		CUtilities::trim(zorder_tokens_vec[i]);		// removes all spaces in string tokens that to the right of each '_'
		variable_indices += zorder_tokens_vec[i];	// the index depth is stored in a left-to-right fashion
													// z_{d,n} means there is a z for every n in d
	}

	// tokenizing the generating distribution part for observed flags
	CUtilities::tokenize(generating_distribution, observed_tokens_vec, observed_flag_delimiter);

	CUtilities::trim_left_right(observed_tokens_vec[0]);
	generating_distribution = observed_tokens_vec[0];	// this is the generating distribution with the parameters in (...)

	// tokenize the parameters here
	CUtilities::tokenize(generating_distribution, parameter_tokens_vec, parameter_index_delimiter);
	generating_distribution = parameter_tokens_vec[0];	// there should be at least one token - leftmost being the name

	SModelGraphNode* node = new SModelGraphNode(variable, generating_distribution);
	_variables_seen_so_far_set.insert(variable);

	if ( observed_tokens_vec.size() > 1 ) {

		if ( observed_tokens_vec[1].compare("observed") == 0 )
			node->set_observed_flag(true);
	}
	node->set_variable_indices(variable_indices);

	// take care while handling the parameters - some may not be numbers i.e. parameters may not be numbers in which case they represent the parent variable
	if ( parameter_tokens_vec.size() > 1 ) {	// if there is an associated parameter - there should be at least 1 due to (...)
		string params = parameter_tokens_vec[1];
		CUtilities::trim(params);

		/////////////////////////////////////////////////////
		// compute a finite state on the full param string //
		//if (b_verbose_global) std::cout << "gen_dist params = " << params << endl;
		populate_node_params(node, params);
	}

	//if (b_verbose_global) std::cout << variable << " [variable_indices = " << variable_indices << "] ~" << generating_distribution << endl;

	//if (b_verbose_global) std::cout << "building graphical model..." << endl;

	map<SModelGraphNode*, list<SModelGraphNode*>*, CModelGraphNodeCompare>::iterator graph_itr;
	graph_itr = _model_graph->find(node);
	if ( graph_itr == _model_graph->end() ) {
		_model_graph->insert( make_pair(node, new list<SModelGraphNode*>) );
	} else {

	}

	//if (b_verbose_global) std::cout << "end of a node ------- " << endl;
	//print_graph();
}

/**
 * This is a private function to populate the node's generating distributions parameter
 */
void CModelGraphBuilder::populate_node_params(SModelGraphNode* node, string param) {
	// compute a finite state machine on the full param string
	size_t i = 0;
	size_t str_len = param.length();
	const char* param_str = param.c_str();
	string seen = "";
	double d = 0; bool is_number;
	list<double> param_numeric_list;	// if this vector's size is > 0 then this will contain the parent distribution's numeric parameters
	list<string> param_string_list;	// else, this contains info about the name of the parent variable and the dimensionality

	stack<char> brace_stack;	// this is not required in this simple parser anyway --should be deprecated

	while ( i < str_len ) {
		switch (param_str[i]) {
		case ',':
			is_number = CUtilities::string_to_number<double>(d, seen, std::dec);
			//if (b_verbose_global) std::cout << "seen = " << seen << " -> " << is_number << " -> " << d << endl;
			if ( is_number )
				param_numeric_list.push_back(d);
			else if ( brace_stack.empty() )	// so essentially if no { is found before , it is should be a parent variable name.
				param_string_list.push_back(seen);
			seen = "";
			break;
		case '{':
			brace_stack.push('{');
			if ( seen.compare("") != 0 ) {
				is_number = CUtilities::string_to_number<double>(d, seen, std::dec);
				//if (b_verbose_global) std::cout << "seen = " << seen << " -> " << is_number << " -> " << d << endl;
				seen = "";
			}
			break;
		case '}':
			brace_stack.pop();
			break;
		case '_':
			//if ( brace_stack.empty() )	// so essentially if no { is found before , it is ok.
										// so if we find something like z_{d,n}, we push "z"
				param_string_list.push_back(seen);

			break;
		default:
			seen += param_str[i];
			break;
		}
		i++;
	}
	is_number = CUtilities::string_to_number<double>(d, seen, std::dec);
	//if (b_verbose_global) std::cout << "seen = " << seen << " -> " << is_number << "->" << d << endl;
	if ( is_number )
		param_numeric_list.push_back(d);
	else if ( brace_stack.empty() )	// so essentially if no { is found before , it is ok.
		param_string_list.push_back(seen);

	if ( param_numeric_list.size() > 0 ) node->set_generating_distribution_numeric_parameter_list(param_numeric_list);
	if ( param_string_list.size() > 0 ) node->set_generating_distribution_string_parameter_list(param_string_list);

	//if (b_verbose_global) std::cout << node->to_string() << endl;
}

/**
 * This function builds the complete graph (populated the dependency links from the initial set of nodes
 * This function actually populates the empty lists corresponding to each node.
 * 		The empty lists were created in the add_node_to_graph function
 */
void CModelGraphBuilder::build_complete_graph() {
	_graphical_model->build_complete_graph();
}


