/*
 * CModelGraphBuilder.h
 *
 *  Created on: May 17, 2010
 *      Author: pdas
 */

#ifndef CMODELGRAPHBUILDER_H_
#define CMODELGRAPHBUILDER_H_

#include "../include/includes.h"
#include "../include/CUtilities.h"
#include "CModelGraph.h"
//using namespace std;

/**
 * The CModelGraphBuilder is a class that attempts to build a graph data structure from a model description file
 * The ModelFileParser class creates an instance of this class and delegates each dependency (a pair of cause-effect nodes)
 * as a std::string to this class.
 *
 * The graph here is represented as a directed graph of hidden or observed variables
 * Note: Instead of building the graph as a list of lists or a std::vector of lists, we use a map of (node, list) representation
 * 		- Such a representation would help us in looking up nodes in the graph in logarithmic or constant time (depending on implementation).
 * 		- Each variable (node) under the map representation must have a unique name (for e.g. alpha, beta etc.)
 */
class CModelGraphBuilder {

	// This is the object representing the graphical model
	CModelGraph* _graphical_model;

	// This map is the graph representing the statistical model -- obtained from CModelGraph object
	std::map<SModelGraphNode*, std::list<SModelGraphNode*>*, CModelGraphNodeCompare> *_model_graph;

	std::set<std::string> _variables_seen_so_far_set;	// this is a set to track which parent variables got added

public:
	CModelGraphBuilder();
	virtual ~CModelGraphBuilder();
	void instantiate_graphical_model();

	// accessors

	inline CModelGraph* get_graphical_model() { return _graphical_model; }

	inline void print_model() {
		_graphical_model->print_graph();
	}

	////////////////////////
	// core api functions //
	void add_node_to_graphical_model(std::string variable, std::string generating_distribution);
	void build_complete_graph();

private:
	void populate_node_params(SModelGraphNode* node, std::string param);

};

#endif /* CMODELGRAPHBUILDER_H_ */
