/*
 * CModelFileSimpleParser.cpp
 *
 *  Created on: May 17, 2010
 *      Author: pdas
 */

#include "CModelFileSimpleParser.h"

using namespace std;

/**
 * The constructor just creates an instance of the graphical model as a graph data structure
 */
CModelFileSimpleParser::CModelFileSimpleParser() {
	// the graph representation of the model is instantiated here
	_graphical_model_builder = new CModelGraphBuilder();
	_graphical_model_builder->instantiate_graphical_model();
}

/**
 * The destructor deletes the graph model
 */
CModelFileSimpleParser::~CModelFileSimpleParser() {
	delete _graphical_model_builder;
}

/**
 * This function just reads in the statistical pseudocode file and builds a
 * graph data-structure representing the model
 */
void CModelFileSimpleParser::parse(string pseudocode_filename) {

	string line;							// the line of the file read by the getline function
	vector<string> tokens_vec;				// the vector of strings that is obtained after tokenizing each line (see CUtilities::tokenize)
	string distribution_delimiter = "~";	// the delimiters used for tokenization

	// some more temporary variables
	string variable_name, distribution_name;


	/*
	 * Note that the pseudocode_filename contains the full path of the filename
	 */
	ifstream fs(pseudocode_filename.c_str());
	if ( fs.is_open() ) {

		while ( !fs.eof() ) {
			getline(fs, line);

			CUtilities::trim_left_right(line);

			if ( !line.length() ) continue;

			tokens_vec.clear();		// typically this vector needs to be cleared before each call to tokenize()
			CUtilities::tokenize(line, tokens_vec, distribution_delimiter);
			// process the tokens here - eg. build the graph "data-structure" corresponding to the graphical model
			// there should be just two tokens after tokenization with "~" as in alpha ~ gam(0.1,1)
			if ( tokens_vec.size() != 2 ) { if (b_verbose_global) std::cout << "This line in the model file is overlooked: " << line << endl; continue; }

			//if (b_verbose_global) std::cout << tokens_vec[0] << endl;
			// parse (trivially) tokens[0] for node variable name
			CUtilities::trim_left_right(tokens_vec[0]);
			//if (b_verbose_global) std::cout << tokens_vec[0] << endl;

			//if (b_verbose_global) std::cout << tokens_vec[1] << endl;
			// parse (non-trivially) tokens[1] for distribution information
			CUtilities::trim_left_right(tokens_vec[1]);
			//if (b_verbose_global) std::cout << tokens_vec[1] << endl;

			_graphical_model_builder->add_node_to_graphical_model(tokens_vec[0], tokens_vec[1]);
		}
		fs.close();
	} else { if (b_verbose_global) std::cout << "Couldn't open " << pseudocode_filename << endl; }

	///////////////////////////////////////////////////
	/// the graph data structure of the model is    ///
	/// fully built at this stage					///
	_graphical_model_builder->build_complete_graph();//
	///////////////////////////////////////////////////


}
