/*
 * CLDASamplingFunctionoid.cpp
 *
 *  Created on: May 21, 2010
 *      Author: pdas
 */

#include "CLDASamplingFunctionoid.h"
#include "../DataSet/CParameterMatrixReaderWriter.h"

using namespace std;

CLDASamplingFunctionoid::CLDASamplingFunctionoid() {
	_concrete_model = new CLDAGibbs();
	_base_model = dynamic_cast<CLDAGeneralModel*>(_concrete_model);
	_likelihood = 0;
}

CLDASamplingFunctionoid::~CLDASamplingFunctionoid() {
	delete _concrete_model;

}

/**
 * This is the steroid that works up the sampler - for as long as we wish :-)
 */
void CLDASamplingFunctionoid::operator()() {

	_prob_density_vec.resize(_concrete_model->_K, 0.0);	// must resize this vector here

	////////////////////////////////////////////////////////////////
	//////////////////////////////// open likelihood file for saving
	string likelihood_filename = _concrete_model->get_model_root_dir() +
			CUtilities::_file_path_separator + "likelihood." + _concrete_model->get_model_operation_mode() + ".txt";
	FILE* fp_likelihood = fopen(likelihood_filename.c_str(), "w");
	if ( !fp_likelihood ) {
		if (b_verbose_global) std::cout << "Cannot open likelihood file " << likelihood_filename << endl;
		return;
	}

	//////////////////////////////////// timing start //////////////
	//time_t tbegin;
	//time(&tbegin);
	//tbegin = time(NULL);
	////////////////////////////////////// timing end //////////////
//#ifdef CUSTOMIZED
	char graphic_art[4] = {'-', '\\', '|', '/'};
//#endif

	////////////////////////////////////////////////////////////////
	//////////////////////////// setting up for iteration parameters
	_concrete_model->print_parameters();	// TODO - remove
	size_t n_gibbs_iter = 0, gibbs_iter_burnin = 0, gibbs_iter_lag = 0;

	n_gibbs_iter = _concrete_model->_n_gibbs_iter;				// <---
	gibbs_iter_burnin = _concrete_model->_gibbs_iter_burnin;	//	   | The iteration params will really depend on how the config files were written
	gibbs_iter_lag = _concrete_model->_gibbs_iter_lag;			// <---
	///////////////////////////////////////////////////////////////////
	///////////////////////////////////////////////////////////////////

	///////////////////////////////////////////////////////////////////
	//////////////////////////////////// run the EM equivalent of gibbs
	if (b_verbose_global) cout << "from lda functionoid operator(): n_gibbs_iter = " << n_gibbs_iter
		 << " burnin = " << gibbs_iter_burnin << " lag = " << gibbs_iter_lag << endl;
	////////////////////////////////////////////// iterate over n steps
	for ( size_t n_steps = 0; n_steps < n_gibbs_iter; ++n_steps ) {

		imputation_step();		// I step : imputation in the LDA context means that you are ascribing a topic ID to z_n for w_n

		if ( ( ((n_steps+1)%gibbs_iter_lag == 0) && ( (n_steps+1) > gibbs_iter_burnin-1) ) || (n_steps == 0) ) {
			posterior_step();	// P step : posterior computation in the usual sense

			_likelihood = _concrete_model->compute_log_likelihood();

			if (b_verbose_global) fprintf(stdout,"Iteration = %4zd .. log likelihood = %16.5lf\n", n_steps+1, _likelihood);
			if (b_verbose_global) fflush(stdout);

			fprintf(fp_likelihood, "%16.6lf\n",  _likelihood);
		}

//#ifdef CUSTOMIZED
		fprintf(stdout, "\r%6.2f percent remaining \t [%c]\r", 100.0*(n_gibbs_iter - (n_steps+1))/n_gibbs_iter, graphic_art[n_steps%4]);
		fflush(stdout);
//#endif

		/*else {
			posterior_step();											//	----
			_likelihood = _concrete_model->compute_log_likelihood();	//		| To be
			fprintf(fp_likelihood, "%16.6lf\n",  _likelihood);			//		| removed
																		//	----
			if ( (n_steps + 1)%50 == 0 ) {fprintf(stdout, "Iteration = %4d .. log likelihood = % 16.6lf\n",  n_steps+1, _likelihood);
			fflush(stdout); }
		}*/
	}
	posterior_step();			// final P step
	cout << endl << "done sampling." << endl;

	if ( _concrete_model->get_model_operation_mode() == "train" ) {
		fprintf(fp_likelihood, "BIC = %16.6lf\n", _concrete_model->get_BIC());
		fprintf(fp_likelihood, "MDL = %16.6lf\n", _concrete_model->get_MDL());
	}
	fclose(fp_likelihood);

	////////////////////////////////////////// timing start ////////////
	//time_t tfinish;
	//time(&tfinish);
	//tfinish = time(NULL);
	//if (b_verbose_global) std::cout << "Total time to process = " << (tfinish-tbegin)/60.0 << " mins" << endl;
	///////////////////////////////////////////// timing end ////////////

	/////////////////////////////////////////////////////////////////////////////////////////////////
	// DIAGNOSTICS
	if ( _concrete_model->get_model_operation_mode() == "train" ) {
		string tw_filename = _concrete_model->get_model_root_dir() +
				CUtilities::_file_path_separator + "topicwords.txt"; // deleted + CUtilities::number_to_string(_concrete_model->_K)
		ofstream ofs(tw_filename.c_str());
		string index_file = _concrete_model->_datum_to_id_dict_filename;
		if (b_verbose_global) std::cout << "reading index file " << index_file << " and writing topics to file " << tw_filename << endl;
		CParameterMatrixReaderWriter pmr;
		pmr.print_topics(_concrete_model->_beta, index_file, 25, ofs);

	} else if ( _concrete_model->get_model_operation_mode() == "test" ) {

		// see CModelTestCPPUnit.cpp
	}
	//////////////////////////////////////////////////////////////////////////////////////////////////
}

/**
 * computes the I (imputation) step for sampling
 */
void CLDASamplingFunctionoid::imputation_step() {
	//if (b_verbose_global) std::cout << "The CLDAfunctionoid is in the Imputation Step... " ;

	///////////////////////////////////////////////////////
	///		resampling methods for hidden variables		///
	if ( _concrete_model->get_model_operation_mode() == "train" )
		resample_z_train();
	if ( _concrete_model->get_model_operation_mode() == "test")
		resample_z_test();
	//if (b_verbose_global) std::cout << "done.." << endl;
}

/**
 * computes the P (posterior) step for sampling
 */
void CLDASamplingFunctionoid::posterior_step() {
	//if (b_verbose_global) std::cout << "The CLDAfunctionoid is in the Posterior step... ";
	if ( _concrete_model->get_model_operation_mode() == "train" ) {
		resample_theta_train();
		resample_beta_train();
	}
	if ( _concrete_model->get_model_operation_mode() == "test" ) {
		resample_theta_test();
		resample_beta_test();
	}

	//resample_alpha();
	//resample_eta();
	//if (b_verbose_global) std::cout << "done.." << endl;
}

/**
 * @brief returns a sample from posterior \\theta having the following generative distribution
 * 		o (shape, scale) \n
 * 		| 				\n
 * 		v 				\n
 * 		o \\alpha ~ Gam(shape, scale) \n
 * 		|							 \n
 * 		v							 \n
 * 		o \\theta ~ DirSym(\\alpha) 	\n
 * 		|							 \n
 * 		v							 \n
 * 		o \\bm{z} ~ Mult(\\theta)	 \n
 * 		|							 \n
 * 		v							Mult	\\beta	sym_dir \\eta	 \n
 * 		o \\bm{w} ~ Mult(\\beta_{z}) <----------o <-------------o		\n
 *
 * In Essence, this function simulates the following loop in one chunk
 * for all z_i
 *	for (int m = 0; m < M; m++) {
 *	    for (int n = 0; n < ptrndata->docs[m]->length; n++) {
 *		// (z_i = z[m][n])
 *		// sample from p(z_i|z_-i, w)
 *		int topic = sampling(m, n);	// m is the document index and n is the nth word in the mth document
 *		z[m][n] = topic;
 *	    }
 *	}
 *
 *	@note: this function is called during training only
 */
void CLDASamplingFunctionoid::resample_z_train() {

	size_t n_docs = _concrete_model->_z_d_n->size();	// no. of documents in the corpus
	size_t n_doc_terms;									// no. of unique terms per doc
	size_t v_id;										// id of the datum in the vocabulary
	size_t v_id_per_doc_count;								// count of the v_id in a particular doc
	size_t topic;										// this is the topic category or topic id
	vector<int>* topic_assignment_vec;

	double alpha = _concrete_model->_alpha;
	double eta = _concrete_model->_eta;

	double K_times_alpha = _concrete_model->_K * alpha;
	double V_times_eta = _concrete_model->_V * eta;

	// iterate over all contexts in which the data is observed
	for ( size_t d = 0; d < n_docs; ++d ) {
		topic_assignment_vec = _concrete_model->_z_d_n->at(d);
		n_doc_terms = _concrete_model->_corpus->_document_vec->at(d)->_datum_id_vec->size();

		// iterate over each datum in the context
		for ( size_t n = 0 ; n < n_doc_terms ; ++n ) {
			topic = topic_assignment_vec->at(n);

			v_id = _concrete_model->_corpus->_document_vec->at(d)->_datum_id_vec->at(n);
			v_id_per_doc_count = _concrete_model->_corpus->_document_vec->at(d)->_datum_id_count_vec->at(n);

			_concrete_model->_z_v_k->at(v_id)->at(topic) -= (1 * v_id_per_doc_count);
			_concrete_model->_z_v_k_total->at(topic) -= (1 * v_id_per_doc_count);
			_concrete_model->_z_d_k->at(d)->at(topic) -= (1 * v_id_per_doc_count);
			_concrete_model->_z_d_k_total->at(d) -= (1 * v_id_per_doc_count);

			// do multinomial sampling via cumulative method
			for (size_t k = 0; k < _concrete_model->_K; ++k) {
				_prob_density_vec[k] =
					((_concrete_model->_z_d_k->at(d)->at(k) + alpha) / (_concrete_model->_z_d_k_total->at(d) + K_times_alpha)) *
					((_concrete_model->_z_v_k->at(v_id)->at(k) + eta) / (_concrete_model->_z_v_k_total->at(k) + V_times_eta));
			}

			topic = sample_single_mult(_prob_density_vec);

			// add newly estimated z_i to count variables
			_concrete_model->_z_v_k->at(v_id)->at(topic) += (1 * v_id_per_doc_count);
			_concrete_model->_z_v_k_total->at(topic) += (1 * v_id_per_doc_count);
			_concrete_model->_z_d_k->at(d)->at(topic) += (1 * v_id_per_doc_count);
			_concrete_model->_z_d_k_total->at(d) += (1 * v_id_per_doc_count);

			// reassign topic category for this datum
			_concrete_model->_z_d_n->at(d)->at(n) = topic;
		}
	}

}

/**
 * This function re-samples beta during training
 */
void CLDASamplingFunctionoid::resample_beta_train() {
	//if (b_verbose_global) std::cout << "p-step beta " ;
	conj_dirsym_mult_posterior_mode(
			_concrete_model->_beta,
			_concrete_model->_z_v_k,
			true,
			_concrete_model->_z_v_k_total,
			_concrete_model->_eta,
			_concrete_model->_V);

}

/**
 * This function resamples theta during training
 */
void CLDASamplingFunctionoid::resample_theta_train() {
	//if (b_verbose_global) std::cout << "p-step theta" ;
	conj_dirsym_mult_posterior_mode(
			_concrete_model->_theta,
			_concrete_model->_z_d_k,
			false,
			_concrete_model->_z_d_k_total,
			_concrete_model->_alpha,
			_concrete_model->_K);
}

/**
 * @brief returns a sample from posterior \\theta having the following generative distribution
 * 		o (shape, scale) \n
 * 		| 				\n
 * 		v 				\n
 * 		o \\alpha ~ Gam(shape, scale) \n
 * 		|							 \n
 * 		v							 \n
 * 		o \\theta ~ DirSym(\\alpha) 	\n
 * 		|							 \n
 * 		v							 \n
 * 		o \\bm{z} ~ Mult(\\theta)	 \n
 * 		|							 \n
 * 		v							Mult	\\beta	sym_dir \\eta	 \n
 * 		o \\bm{w} ~ Mult(\\beta_{z}) <----------o <-------------o		\n
 *
 * In Essence, this function simulates the following loop in one chunk
 * for all z_i
 *	for (int m = 0; m < M; m++) {
 *	    for (int n = 0; n < ptrndata->docs[m]->length; n++) {
 *		// (z_i = z[m][n])
 *		// sample from p(z_i|z_-i, w)
 *		int topic = sampling(m, n);	// m is the document index and n is the nth word in the mth document
 *		z[m][n] = topic;
 *	    }
 *	}
 *
 *	@note: this function is called during training only
 */
void CLDASamplingFunctionoid::resample_z_test() {

	//if (b_verbose_global) std::cout << "i-step z_test in test... " ;

	size_t n_docs_test = _concrete_model->_z_d_n_test->size();	// no. of documents in the corpus
	size_t n_doc_terms_test;									// no. of unique terms per doc
	size_t v_id_test;											// id of the test datum in the vocabulary
	size_t v_id_test_per_doc_count;								// count of the v_id in a particular doc
	size_t topic;												// this is the topic category or topic id
	vector<int>* topic_assignment_vec_test;
	size_t term_id_test;

	double alpha = _concrete_model->_alpha;
	double eta = _concrete_model->_eta;

	double K_times_alpha = _concrete_model->_K * alpha;
	double V_times_eta = _concrete_model->_V * eta;

	assert(_concrete_model->_z_v_k != 0);						// an important assert to keep anyway

	map<size_t, size_t>::iterator m_itr;

	// iterate over all contexts in which the data is observed
	for ( size_t d = 0; d < n_docs_test; ++d ) {

		topic_assignment_vec_test = _concrete_model->_z_d_n_test->at(d);
		n_doc_terms_test = _concrete_model->_corpus_test->_document_vec->at(d)->_num_of_unique_datum;

		// iterate over each datum in the context

		for ( size_t n = 0 ; n < n_doc_terms_test ; ++n ) {

			topic = topic_assignment_vec_test->at(n);
			term_id_test = _concrete_model->_corpus_test->_document_vec->at(d)->_datum_id_vec->at(n);

			// we have test term ids like 10001, 10002, etc. which are mapped to 0, 1, etc. so that the _V_test is both 0 indexed, uniformly incremented by 1
			// and _V_test strictly <= _V
			m_itr = _concrete_model->get_test_datum_id_to_sequential_index_id_map().find(term_id_test);
			if ( m_itr == _concrete_model->get_test_datum_id_to_sequential_index_id_map().end() ) {
				if (b_verbose_global) std::cout << "this shouldn't happen for term id " << term_id_test << endl;
				exit(APPLICATIONEXIT);
			}
			v_id_test = m_itr->second;


			//v_id = _concrete_model->get_test_datum_id_to_sequential_index_id_map()[_concrete_model->_corpus_test->_document_vec->at(d)->_datum_id_vec->at(n)];
			v_id_test_per_doc_count = _concrete_model->_corpus_test->_document_vec->at(d)->_datum_id_count_vec->at(n);

			_concrete_model->_z_v_k_test->at(v_id_test)->at(topic) -= (1 * v_id_test_per_doc_count);
			_concrete_model->_z_v_k_total_test->at(topic) -= (1 * v_id_test_per_doc_count);
			_concrete_model->_z_d_k_test->at(d)->at(topic) -= (1 * v_id_test_per_doc_count);
			_concrete_model->_z_d_k_total_test->at(d) -= (1 * v_id_test_per_doc_count);

			//if (b_verbose_global) std::cout << "<topic = " << topic << " term_id_test = " << term_id_test << " v_id_test = " << v_id_test << endl; // TODO -remove

			// do multinomial sampling via cumulative method
			for (size_t k = 0; k < _concrete_model->_K; ++k) {
				_prob_density_vec[k] =
					((_concrete_model->_z_d_k_test->at(d)->at(k) + alpha) / (_concrete_model->_z_d_k_total_test->at(d) + K_times_alpha)) *
					((_concrete_model->_z_v_k->at(v_id_test)->at(k) + _concrete_model->_z_v_k_test->at(v_id_test)->at(k) + eta) / (_concrete_model->_z_v_k_total->at(k) + _concrete_model->_z_v_k_total_test->at(k) + V_times_eta));
			}

			topic = sample_single_mult(_prob_density_vec);

			// add newly estimated z_i to count variables
			_concrete_model->_z_v_k_test->at(v_id_test)->at(topic) += (1 * v_id_test_per_doc_count);
			_concrete_model->_z_v_k_total_test->at(topic) += (1 * v_id_test_per_doc_count);
			_concrete_model->_z_d_k_test->at(d)->at(topic) += (1 * v_id_test_per_doc_count);
			_concrete_model->_z_d_k_total_test->at(d) += (1 * v_id_test_per_doc_count);

			// reassign topic category for this datum
			_concrete_model->_z_d_n_test->at(d)->at(n) = topic;

			//if (b_verbose_global) std::cout << ">topic = " << topic << " term_id_test = " << term_id_test << " v_id_test = " << v_id_test << endl; // TODO -remove
		}
		//if (b_verbose_global) std::cout << "done with test doc_id " << d << endl;
	}

	// this function is passing
	//if (b_verbose_global) std::cout << " done." << endl;
}

/**
 * This function re-samples beta for testing
 */
void CLDASamplingFunctionoid::resample_beta_test() {
	//if (b_verbose_global) std::cout << "p-step beta in test... " ;
	vector<vector<vector<int>* >*> mat;
	vector<vector<int>* > mat_sum;
	mat.push_back(_concrete_model->_z_v_k);
	mat.push_back(_concrete_model->_z_v_k_test);
	mat_sum.push_back(_concrete_model->_z_v_k_total);
	mat_sum.push_back(_concrete_model->_z_v_k_total_test);

	conj_dirsym_mult_posterior_mode(
			_concrete_model->_beta_test,
			&mat,
			true,
			&mat_sum,
			_concrete_model->_eta,
			_concrete_model->_V);
	//if (b_verbose_global) std::cout << " done." << endl;
}

/**
 * This function resamples theta
 */
void CLDASamplingFunctionoid::resample_theta_test() {
	//if (b_verbose_global) std::cout << "p-step theta in test... " ;
	conj_dirsym_mult_posterior_mode(
			_concrete_model->_theta_test,
			_concrete_model->_z_d_k_test,
			false,
			_concrete_model->_z_d_k_total_test,
			_concrete_model->_alpha,
			_concrete_model->_K);
	//if (b_verbose_global) std::cout << " done." << endl;
}

/**
 * This function resamples prior alpha for topic proportions
 */
void CLDASamplingFunctionoid::resample_alpha() {

}

/**
 * This function resamples prior eta for word proportions in topics
 */
void CLDASamplingFunctionoid::resample_eta() {

}

