/*
 * CModelSamplingFunctionoid.h
 *
 *  Created on: May 21, 2010
 *      Author: pdas
 */

#ifndef CMODELSAMPLINGFUNCTIONOID_H_
#define CMODELSAMPLINGFUNCTIONOID_H_

#include "../GraphicalModels/CLDAGeneralModel.h"
#include "AModelSamplingFunctionoid.h"

/**
 * @brief To-do documentation
 */
class CModelSamplingFunctionoid : public AModelSamplingFunctionoid {

protected:
	/*!
	 * A structure used by normalize_vector_double()
	 */
	struct SNormFunctionObj {
		double _sum;
		SNormFunctionObj(double sum) : _sum(sum) {}
		double operator() (double n) const { return n/_sum; }
	};

	/*!
	 * a boolean indicating whether the verbose output is set to true. By default it is set to true
	 */
	bool _is_verbose;

public:
	CModelSamplingFunctionoid();
	virtual ~CModelSamplingFunctionoid();
	/**
	 *
	 */
	virtual inline void set_verbose_mode(bool is_verbose) { _is_verbose = is_verbose; }
	/**
	 *
	 */
	virtual inline CLDAGeneralModel* get_concrete_model() { if (b_verbose_global) std::cout << "You should derive your own C<YourModel>SamplingFunctionoid from CModelSamplingFunctionoid" << std::endl; return 0; }
	/**
	 *
	 */
	virtual inline void operator()() { imputation_step(); posterior_step(); }
	/**
	 *
	 */
	virtual double get_final_model_log_likelihood() { if (b_verbose_global) std::cout << "returning likelihood from base (shouldn't happen) = " ; return 0; }

private:
	/**
	 *
	 */
	virtual void imputation_step() { if (b_verbose_global) std::cout << "The base functionoid doesn't do any imputation" << std::endl; }
	/**
	 *
	 */
	virtual void posterior_step() { if (b_verbose_global) std::cout << "The base functionoid doesn't do any posterior calculations either" << std::endl; }

protected:

	/////////////////////////////////////////////////////
	///////// HBI SAMPLING FUNCTIONS BLOCK //////////////
	/**
	 *
	 */
	double gam_symdir_posterior_mode(double shape, double scale, double old_prior, double D);
	/**
	 *
	 */
	int sample_single_mult(std::vector<double>& pdf);
	/**
	 *
	 */
	void normalize_vector_double(std::vector<double>* v);

	/////////////////////////////////////////////////////////////////////////////////////
	///////// HBI POSTERIOR MODE FINDING FOR CONJs IN THE POSTERIOR P-STEP //////////////
	/**
	 *
	 * @brief returns a sample from posterior \\theta having the following generative distribution
	 * 		o \\alpha
	 * 		|
	 * 		v
	 * 		o \\theta ~ DirSym(\\alpha)
	 * 		|
	 * 		v
	 * 		o \\bm{z} ~ Mult(\\theta)
	 *
	 * This function handles the following pseudocode for posterior mode finding\n
	 * CASE 1:
	 * \code
	 * resample_theta() {
	 *  for (size_t d = 0; d < _D; ++d ) {
	 *	 	for (size_t k = 0; k < _K; ++k) {
	 *			theta(d,k) = (_z_d_k[d][k] + _alpha) / (_z_d_k_total[d] + _K * _alpha);
	 *		}
	 *	}
	 * }
	 * \endcode
	 * CASE 2:
	 * \code
	 * resample_phi() {
	 *   for (size_t k = 0; k < _K; ++k) {
	 *		for (size_t v = 0; v < _V; ++v) {
	 *	    	phi(k,j) = (_z_v_k[v][k] + _beta) / (_z_v_k_total[k] + _V * _beta);
	 *		}
	 *	  }
	 * }
	 * \endcode
	 * For the second case, the count_matrix is transposed since it indicates how many times has topic k been associated with the datum vs.
	 * By default, bool count_matrix_transposed is set to false
	 *
	 * @note: the matrices mode_variable_matrix and count_matrix are full matrices and not sparse matrices

	 * @param mode_variable_matrix is a pointer to std::vector<std::vector<double>* >
	 * @param count_matrix is a pointer to std::vector<std::vector<int>* >
	 * @param count_matrix_transposed_bool is a boolean indicating whether the count matrix is transposed or not
	 * @param count_matrix_sum is a pointer to std::vector<int>
	 * @param sym_prior is a double representing the value of the symmetric prior
	 * @param n_categories is an unsigned int representing the value of the dimensionality of the posterior distribution
	 *
	 * @note
	 * This function updates the values in the mode_variable_matrix
	 */
	void conj_dirsym_mult_posterior_mode(
			std::vector<std::vector<double>* >* mode_variable_matrix,
			std::vector<std::vector<int>* >* count_matrix,
			bool count_matrix_transposed_bool,
			std::vector<int>* count_matrix_sum,
			double sym_prior,
			size_t n_categories);

	/**
	 * @brief Overloaded CModelSamplingFunctionoid::conj_dirsym_mult_posterior_mode() with modified parameters
	 * vec_of_count_matrices and vec_of_count_matrix_sums
	 *
	 * @param mode_variable_matrix is a pointer to std::vector<std::vector<double>* >
	 * @param vec_of_count_matrices is a pointer to std::vector<std::vector<std::vector<int>* >*>
	 * @param count_matrix_transposed_bool is a boolean indicating whether the count matrix is transposed or not
	 * @param vec_of_count_matrix_sums is a pointer to std::vector<std::vector<int>*>
	 * @param sym_prior is a double representing the value of the symmetric prior
	 * @param n_categories is an unsigned int representing the value of the dimensionality of the posterior distribution
	 *
	 * @note
	 * This function updates the values in the mode_variable_matrix
	 */
	void conj_dirsym_mult_posterior_mode(
			std::vector<std::vector<double>* >* mode_variable_matrix,
			std::vector<std::vector<std::vector<int>* >*>* vec_of_count_matrices,
			bool count_matrix_transposed_bool,
			std::vector<std::vector<int>*>* vec_of_count_matrix_sums,
			double sym_prior,
			size_t n_categories);

#ifdef UNIX
	//////////////////////////////////////////////////////
	////////// HBC SAMPLING FUNCTIONS BLOCK //////////////
	/**
	 * \defgroup HBC_SAMPLING_FUNCTIONS_BLOCK
	 */
	/*@{*/
	double gammaln(double x);
	double digamma(double x);
	double trigamma(double x);
	double invdigamma(double x);

	void advnst(long k);
	double genbet(double aa,double bb);
	double genchi(double df);
	double genexp(double av);
	double genf(double dfn, double dfd);
	double gengam(double a,double r);
	void genmn(double *parm,double *x,double *work);
	void genmul(long n,double *p,long ncat,long *ix);
	double gennch(double df,double xnonc);
	double gennf(double dfn, double dfd, double xnonc);
	double gennor(double av,double sd);
	void genprm(long *iarray,int larray);
	double genunf(double low,double high);
	void getsd(long *iseed1,long *iseed2);
	void gscgn(long getset,long *g);
	long ignbin(long n,double pp);
	long ignnbn(long n,double p);
	long ignlgi(void);
	long ignpoi(double mu);
	long ignuin(long low,long high);
	void initgn(long isdtyp);
	long mltmod(long a,long s,long m);
	void phrtsd(char* phrase,long* seed1,long* seed2);
	double ranf(void);
	void setall(long iseed1,long iseed2);
	void setant(long qvalue);
	void setgmn(double *meanv,double *covm,long p,double *parm);
	void setsd(long iseed1,long iseed2);
	double sexpo(void);
	double sgamma(double a);
	double snorm(void);
	/*@}*/
	//////////////////////////////////////////////////////
#endif
};

#endif /* CMODELSAMPLINGFUNCTIONOID_H_ */
