/*
 * CModelTestCPPUnit.cpp
 *
 *  Created on: Jul 14, 2010
 *      Author: pdas
 */

#include "CModelTestCPPUnit.h"

using namespace std;

CModelTestCPPUnit::CModelTestCPPUnit(string model_description_file, string model_compiled_file, string config_file) {
	/////////////////////////////////////////////////////////// create the parser
	CModelFileSimpleParser *parser = new CModelFileSimpleParser();
	parser->parse(model_description_file);	// hard-coded now needs to be parameterized

	CPrinter prn;

	//////////////////////////////////////////////////////////////////////////////
	////////////////////////////// From parser -> model builder -> graphical model
	//////////////////////////////////////////////////////////////////////////////
	CModelGraphBuilder* _graphical_model_builder = parser->get_graphical_model_builder();
	if (b_verbose_global) std::cout << "printing FULL model..." << endl;
	CModelGraph* model_graph = _graphical_model_builder->get_graphical_model();
	if (b_verbose_global) prn.print_graphical_model(model_graph);

	SForest* forest = model_graph->get_model_forest_representation();
	if (b_verbose_global) std::cout << "printing level_order_bfs forest on the original model..." << endl;
	if (b_verbose_global) prn.print_graphical_model_forest(forest);

	if (b_verbose_global) std::cout << "printing encoded model..." << endl;
	string matched_model_name = "";
	matched_model_name = model_graph->match_compiled_template(model_compiled_file);	// hardcoded now
	string matched_model_bytecode = model_graph->get_bytecode();

	CUtilities::to_lower(matched_model_name);
	if ( matched_model_name == "" ) matched_model_name = "unrecognized";

	if (b_verbose_global) std::cout << "matched model name = " << matched_model_name << " = " << matched_model_bytecode << endl << endl;
	delete parser;

	//////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////// Let the appropriate model take over now
	//////////////////////////////////////////////////////////////////////////////
	vector<string> cmd_arg_vec;

	CCommandFactory* command_factory = CCommandFactory::get_instance();
	CCommandHandler* command_handler = command_factory->get_command_handler( matched_model_name );

	const size_t num_config_files = 1;
	//string config_files[num_config_files] = {"YAH-LDA-test-config03.txt", "YAH-LDA-test-config05.txt"};
	//string config_files[1] = {"YAH-LDA-test-config05.txt"};
	string config_files[1] = config_file;
	static CParameterMatrixReaderWriter pmr;	// !!!IMPORTANT - DECLARE THIS AS STATIC WHILE TESTING OR VALIDATING

	for ( size_t i = 0; i < num_config_files; ++i ) {

		cmd_arg_vec.clear();
		cmd_arg_vec.push_back("test");
		cmd_arg_vec.push_back(config_files[i]);	// hardcoded now should be parameterized

		if ( command_handler ) {
			command_handler->set_model_bytecode( matched_model_bytecode );
			command_handler->set_arguments( cmd_arg_vec );
			command_handler->handle_command();

			CLDAGeneralModel* model = command_handler->get_command()->get_model();

			vector<vector<int>* > *z_d_n_test, *z_d_k_test;
			z_d_n_test = model->get_test_doc_datum_topic_assignment_vectors();
			z_d_k_test = model->get_test_doc_topic_assignment_vectors();
			if ( z_d_n_test && z_d_k_test ) {
				if (b_verbose_global) std::cout << "dimensions of z_d_n_test are: " << endl;
				for ( size_t d = 0; d < z_d_n_test->size(); ++d ) {
					if (b_verbose_global) std::cout << "1 x " << z_d_n_test->at(d)->size() << endl;
					if (b_verbose_global) std::cout << "Best topic for doc = " << d << " is " << CUtilities::get_max_topic_index(*(z_d_n_test->at(d))) << endl;
					if (b_verbose_global) std::cout << "Best topic for doc = " << d << " (from _z_d_k_test) is " << CUtilities::get_max_index_from_vector<int>( *(z_d_k_test->at(d)) ) << endl;

					/////////////////////////////////////////////////////////////////////////////////////////////////////////
					//////////////////////////////////////////////////////////////////////////////////////////// GLOBAL THEME
					vector<string> expanded_topic_multinomial_string_vec =
							pmr.expand_topic_multinomial_for_entire_corpus(
									model->get_train_saved_topic_multinomial_filename(),
									CUtilities::get_max_topic_index( *(z_d_n_test->at(d)) ),
									model->get_datum_to_id_dictionary_filename(),
									15,
									model->get_model_dump_mode()
							);
					if (b_verbose_global) std::cout << "This (unseen) document belongs to global theme represented by: ";
					for ( size_t i = 0; i < expanded_topic_multinomial_string_vec.size(); ++i )
						if (b_verbose_global) std::cout << expanded_topic_multinomial_string_vec[i] << " ";
					if (b_verbose_global) std::cout << endl;
					/////////////////////////////////////////////////////////////////////////////////////////////////////////
					///////////////////////////////////////////////////////////////////////////////////////////// LOCAL THEME
					vector<string> expanded_topic_multinomial_string_for_this_doc_vec =
							pmr.expand_topic_multinomial_for_this_document(
									model->get_train_saved_topic_multinomial_filename(),
									model->get_test_doc_topic_proportion_multinomials()->at(d),
									model->get_datum_to_id_dictionary_filename(),
									15,
									model->get_model_dump_mode()
							);
					if (b_verbose_global) std::cout << "This (unseen) document belongs to local theme represented by: ";
					for ( size_t i = 0; i < expanded_topic_multinomial_string_for_this_doc_vec.size(); ++i )
						if (b_verbose_global) std::cout << expanded_topic_multinomial_string_for_this_doc_vec[i] << " ";
					if (b_verbose_global) std::cout << endl;
					//SCorpus* corpus_test = model->get_test_corpus();
					//if (b_verbose_global) std::cout << "Corpus size is : " << corpus_test->_document_vec->size() << endl;
				}
			}
		} else {
			if (b_verbose_global) std::cout << "Bad model!!! No doughnuts for you" << endl;
		}
	}

}

CModelTestCPPUnit::~CModelTestCPPUnit() {
	// TODO Auto-generated destructor stub
}
