/*
 * CModelValidateCPPUnit.cpp
 *
 *  Created on: Jul 14, 2010
 *      Author: pdas
 */

#include "CModelValidateCPPUnit.h"
using namespace std;

CModelValidateCPPUnit::CModelValidateCPPUnit(string model_description_file, string model_compiled_file, string config_file) {

	/////////////////////////////////////////////////////////// create the parser
	CModelFileSimpleParser *parser = new CModelFileSimpleParser();
	parser->parse(model_description_file);	// hard-coded now needs to be parameterized

	CPrinter prn;

	//////////////////////////////////////////////////////////////////////////////
	////////////////////////////// From parser -> model builder -> graphical model
	//////////////////////////////////////////////////////////////////////////////
	CModelGraphBuilder* _graphical_model_builder = parser->get_graphical_model_builder();
	if (b_verbose_global) std::cout << "printing FULL model..." << endl;
	CModelGraph* model_graph = _graphical_model_builder->get_graphical_model();
	if (b_verbose_global) prn.print_graphical_model(model_graph);

	SForest* forest = model_graph->get_model_forest_representation();
	if (b_verbose_global) std::cout << "printing level_order_bfs forest on the original model..." << endl;
	if (b_verbose_global) prn.print_graphical_model_forest(forest);

	if (b_verbose_global) std::cout << "printing encoded model..." << endl;
	string matched_model_name = "";
	matched_model_name = model_graph->match_compiled_template(model_compiled_file);	// hardcoded now
	string matched_model_bytecode = model_graph->get_bytecode();

	CUtilities::to_lower(matched_model_name);
	if ( matched_model_name == "" ) matched_model_name = "unrecognized";

	if (b_verbose_global) std::cout << "matched model name = " << matched_model_name << " = " << matched_model_bytecode << endl << endl;
	delete parser;

	//////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////// Let the appropriate model take over now
	//////////////////////////////////////////////////////////////////////////////
	vector<string> cmd_arg_vec;
//#define CROSSVAL
#ifdef CROSSVAL
	bool do_cross_validation;
	do_cross_validation = true;
	if ( do_cross_validation && (matched_model_name != "unrecognized") ) {
		cmd_arg_vec.push_back("cross-validate");
		cmd_arg_vec.push_back(config_file);	// "YAH-LDA-crossval-config-sequential.txt" hardcoded now should be parameterized
		matched_model_name += "-cross-validate";
	}
#endif

#define CROSSVALONLINE
#ifdef CROSSVALONLINE
	bool do_cross_validation;
	do_cross_validation = true;
	if ( do_cross_validation && (matched_model_name != "unrecognized") ) {
		cmd_arg_vec.push_back("cross-validate");
		cmd_arg_vec.push_back(config_file);
		matched_model_name += "-cross-validate-online";
	}
#endif

	CCommandFactory* command_factory = CCommandFactory::get_instance();
	CCommandHandler* command_handler = command_factory->get_command_handler( matched_model_name );

	if ( command_handler ) {
		command_handler->set_model_bytecode( matched_model_bytecode );
		command_handler->set_arguments( cmd_arg_vec );
		command_handler->handle_command();
	} else {
		if (b_verbose_global) std::cout << "Bad model!!! No doughnuts for you" << endl;
	}

}

CModelValidateCPPUnit::~CModelValidateCPPUnit() {
	// TODO Auto-generated destructor stub
}
